from typing import Any, Dict, List, Optional

from django.db import Error, models, transaction
from django.forms import ValidationError

from smoothglue.duplication.exceptions import DuplicationError


def unassignable_field(field):
    """
    Checks to see if a field on a model is assignable based on our rules

    Args:
        field: Django field we are checking the properties of

    Returns:
        bool: Whether or not the field can be assigned
    """
    return getattr(field, "primary_key", False) or isinstance(
        field, (models.ManyToManyField, models.OneToOneField)
    )


def check_fields(original, included_fields, field_mapper):
    """
    Checks to see if the provided fields are suitable for duplication

    Args:
        original: The original object instance we are copying from
        included_fields: Fields we want to directly copy
        field_mapper: Fields we define the values of

    Returns:
        bool: Whether or not the provided fields are valid
    """
    field_names = []
    for field in original._meta.get_fields():
        field_names.append(field.name)
        if unassignable_field(field) and (
            field.name in included_fields or field.name in field_mapper.keys()
        ):
            return False
    valid_included_fields = set(included_fields).issubset(field_names)
    valid_field_mapper = set(field_mapper.keys()).issubset(field_names)
    field_conflicts = list(set(field_mapper.keys()) & set(included_fields))
    return valid_included_fields and valid_field_mapper and len(field_conflicts) == 0


def setup_fields_to_duplicate(
    original: models.Model,
    include_children: bool,
    included_fields: Optional[List[str]],
    excluded_fields: Optional[List[str]],
    field_mapper: Optional[Dict[str, Any]],
):
    """
    Cleans/sets up the fields we are going to be duplicating for the original object.

    Args:
        original: The original object instance we are copying from
        include_children (bool): Boolean whether we want to add
                                and duplicate the children relationships as well.
        included_fields(list[str] | None): Fields we want to directly copy, defaults to all fields
        excluded_fields(list[str] | None): Fields we want to exclude from copy.
        field_mapper(dict[str, Any] | None): Fields we define the values of.

    Returns:
        included_fields, field_mapper: The cleaned version of the provided arguments
    """
    if field_mapper is None:
        field_mapper = {}
    if included_fields is None:
        auto_included_fields = (
            original._meta.get_fields() if include_children else original._meta.fields
        )  # get_fields() includes relationships, fields is only what is directly on the model.

        included_fields = [
            field.name
            for field in auto_included_fields
            if not unassignable_field(field) and field.name not in field_mapper.keys()
        ]
    if excluded_fields is not None:
        included_fields = list(set(included_fields) - set(excluded_fields))

    return included_fields, field_mapper


def get_parent_field_name(original: models.Model, child_model: models.Model):
    """
    Gets the name for the field that represents the parent from it's child relationship

    Args:
        original (models.Model): The original object instance we are copying from
        child_model (models.Model): The child model in the relationship

    Returns:
        str | None: The related name for the parent on the child model, if one is found.
    """
    for field in child_model._meta.get_fields():
        if isinstance(field, models.ForeignKey) and isinstance(
            original, field.related_model
        ):  # Should we add one-2-one here too?
            return field.name
    return None


def duplicate_children(
    duplicated_object: models.Model, one_to_many_relationship_list: List[Dict[str, Any]]
):
    """
    Loops through and duplicates the children from the original object
    and swaps the parent to the duplicate of the original.

    Args:
        duplicated_object (models.Model): The newly duplicated parent
        one_to_many_relationship_list (List[Dict[str, Any]]):
            List of dictionaries containing the related name and queryset of children
    """
    for related_object in one_to_many_relationship_list:
        child_field_mapper = {related_object["parent_field"]: duplicated_object}
        for child_object in related_object.get("related_query", []):
            new_child_object = duplicate_object(
                child_object, True, field_mapper=child_field_mapper
            )
            new_child_object.save()


def duplicate_object(
    original: models.Model,
    include_children: bool,
    included_fields: Optional[List[str]] = None,
    excluded_fields: Optional[List[str]] = None,
    field_mapper: Optional[Dict[str, Any]] = None,
):
    """
    Creates a new object instance based on an original object to duplicate from
    and defined fields.

    Args:
        original: The original object instance we are copying from
        include_children (bool): Boolean whether we want to add
                                and duplicate the children relationships as well.
        included_fields(list[str] | None): Fields we want to directly copy, defaults to all fields
        excluded_fields(list[str] | None): Fields we want to exclude from copy.
        field_mapper(dict[str, Any] | None): Fields we define the values of.

    Returns:
        new_object | None: The newly created instance or None if an issue is found
    """
    copied_fields = {}
    included_fields, field_mapper = setup_fields_to_duplicate(
        original, include_children, included_fields, excluded_fields, field_mapper
    )

    valid_fields = check_fields(original, included_fields, field_mapper)
    if valid_fields:
        one_to_many_relationship_list = []
        for field in included_fields:
            field_instance = original._meta.get_field(field)
            if field_instance.one_to_many:
                parent_field = get_parent_field_name(
                    original, field_instance.related_model
                )
                if parent_field:
                    one_to_many_relationship_list.append(
                        {
                            "parent_field": parent_field,
                            "related_query": getattr(original, field).all(),
                        }
                    )
            else:
                copied_fields[field] = getattr(original, field)
        with transaction.atomic():
            try:
                new_object = original._meta.model(**{**copied_fields, **field_mapper})
                new_object.save()
            except (Error, ValidationError, ValueError) as ex:
                raise DuplicationError(f"Error during duplication: {ex.message}")

            if include_children:
                duplicate_children(new_object, one_to_many_relationship_list)

        return new_object
    raise DuplicationError("Invalid fields provided for duplication.")



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from rest_framework import serializers

from smoothglue.authentication.models import (
    OrganizationCategory,
    OrganizationMember,
    PlatformOrganization,
)
from smoothglue.core.serializers.abstract import DefaultSerializer


class GroupSerializer(serializers.ModelSerializer):
    class Meta:
        model = Group
        fields = ["id", "name"]


class OrganizationCategorySerializer(DefaultSerializer):
    class Meta:
        model = OrganizationCategory
        fields = ["id", "name", "description", "data"]


class PlatformOrganizationSerializer(DefaultSerializer):
    class Meta:
        model = PlatformOrganization
        fields = [
            "id",
            "name",
            "child_organizations",
            "group",
            "abbreviation",
            "org_category",
            "data",
        ]

    def get_fields(self):
        fields = super().get_fields()
        fields["child_organizations"] = PlatformOrganizationSerializer(
            many=True, read_only=True
        )
        return fields


# pylint: disable=W0223
class UserDataSerializer(serializers.Serializer):
    """
    Custom Serializer for JSONField data schema. This will allow the openAPI
    describe what are acceptable for the 'data' JSON format.
    """

    rank = serializers.CharField(required=False)
    usercertificate = serializers.CharField(required=False)
    affiliation = serializers.CharField(required=False)

    def to_representation(self, instance):
        return instance


class PlatformUserSerializer(DefaultSerializer):
    organizations = serializers.PrimaryKeyRelatedField(
        queryset=PlatformOrganization.objects.all(),
        required=False,
        allow_null=False,
        many=True,
    )
    groups = GroupSerializer(read_only=True, many=True)

    class Meta:
        model = get_user_model()
        fields = [
            "id",
            "username",
            "first_name",
            "last_name",
            "email",
            "data",
            "organizations",
            "groups",
            "user_image",
            "is_superuser",
        ]
        extra_kwargs = {
            "organizations": {"required": False},
        }

    def update(self, instance, validated_data):
        org_ids = validated_data.pop("organizations", None)
        instance = super().update(instance, validated_data)

        if org_ids is not None:
            OrganizationMember.objects.filter(platform_user=instance).delete()
            for org in org_ids:
                OrganizationMember.objects.create(
                    platform_user=instance,
                    platform_organization=org,
                    created_by=instance,
                    updated_by=instance,
                )

        return instance

    def create(self, validated_data):
        org_ids = validated_data.pop("organizations", None)
        instance = super().create(validated_data)

        if org_ids is not None:
            for org in org_ids:
                OrganizationMember.objects.create(
                    platform_user=instance,
                    platform_organization=org,
                    created_by=instance,
                    updated_by=instance,
                )

        return instance


class SimplePlatformOrganizationSerializer(DefaultSerializer):
    class Meta:
        model = PlatformOrganization
        fields = [
            "id",
            "name",
        ]


class SimplePlatformUserSerializer(DefaultSerializer):
    class Meta:
        model = get_user_model()
        fields = ["id", "email", "first_name", "last_name", "username"]


class OrganizationMemberSerializer(DefaultSerializer):

    platform_organization = SimplePlatformOrganizationSerializer()
    platform_user = SimplePlatformUserSerializer()

    class Meta:
        model = OrganizationMember
        fields = ["id", "platform_organization", "platform_user"]



from collections import OrderedDict

from rest_framework.serializers import BindingDict


class ModifiedTrieNode:  # pylint: disable=too-few-public-methods
    def __init__(self, value=None, end_of_field=False, fields_dict=None):
        self.value = value
        self.children = {}
        self.end = end_of_field
        self.fields_dict = fields_dict


class ModifiedTrie:
    """
    Recieves a list of strings, each ele in list will be split by
    a '.' and inserted into the trie. The trie will be used to store
    requested fields, eliminate duplicates of the requested fields
    and allow for removing fields that differ between database
    fields and requested fields
    """

    def __init__(self):
        self.__root = ModifiedTrieNode()

    def insert(self, requested_field_section):
        current = self.__root
        for field in requested_field_section.split("."):
            current = current.children.setdefault(field, ModifiedTrieNode(field))
        current.end = True

    def filter_requested_fields(self, fields_dict: BindingDict):
        """
        Traverse tree (breadth) and at each level, compare level
        against the database fields and remove differences. While
        traversing, add node and children(unless node is end) to
        queue. Also pass the database fields
        """

        self.__root.fields_dict = fields_dict  # self.fields
        is_root = True  # accessing fields is different for top level vs nested

        field_queue = [self.__root]
        while field_queue:
            current = field_queue.pop(0)

            if not current.end:  # this stops from traversing lower
                if is_root:
                    default_fields = OrderedDict(current.fields_dict).keys()
                    is_root = False
                elif hasattr(current.fields_dict, "get_fields"):
                    default_fields = current.fields_dict.get_fields().keys()
                else:
                    # a NestedBoundField that doesn't render in time,
                    # i.e. request_allocations/?fields=fuel_request.asset.fuel_requests.id
                    continue

                self.compare_children_to_default_fields(
                    current, default_fields, field_queue
                )

    def compare_children_to_default_fields(
        self, current: ModifiedTrieNode, default_fields, field_queue
    ):
        currents_children = []
        actual_children_field_container = current.fields_dict.fields
        for node in current.children.values():
            if node.value in default_fields:
                node.fields_dict = actual_children_field_container[str(node.value)]

                field_queue.append(node)
                currents_children.append(node.value)

        for field_name in default_fields - set(currents_children):
            current.fields_dict.fields.pop(field_name)



from rest_framework import serializers

from smoothglue.core.models import AppInfo, Classification


class ClassificationSerializer(serializers.Serializer):
    """
    Serializer for the Classification model.
    Serializes the 'level' and 'message' fields.
    """

    level = serializers.CharField(read_only=True)
    message = serializers.CharField(read_only=True)

    def update(self, instance, validated_data):
        """
        Update and return an existing Classification instance, given the validated data.

        Args:
            instance: The existing Classification instance.
            validated_data: The validated data to update the instance with.

        Returns:
            The updated Classification instance.
        """
        instance.level = validated_data.get("level", instance.level)
        instance.message = validated_data.get("message", instance.message)
        return instance

    def create(self, validated_data):
        """
        Create and return a new Classification instance, given the validated data.

        Args:
            validated_data: The validated data to create the instance with.

        Returns:
            The newly created Classification instance.
        """
        return Classification(**validated_data)


class AppInfoSerializer(serializers.Serializer):
    """
    Serializer for the AppInfo model.
    Serializes the 'classification' field using ClassificationSerializer.
    """

    classification = ClassificationSerializer(read_only=True)

    def update(self, instance, validated_data):
        """
        Update and return an existing AppInfo instance, given the validated data.

        Args:
            instance: The existing AppInfo instance.
            validated_data: The validated data to update the instance with.

        Returns:
            The updated AppInfo instance.
        """
        instance.classification = validated_data.get(
            "classification", instance.classification
        )
        return instance

    def create(self, validated_data):
        """
        Create and return a new AppInfo instance, given the validated data.

        Args:
            validated_data: The validated data to create the instance with.

        Returns:
            The newly created AppInfo instance.
        """
        return AppInfo(**validated_data)



import uuid

from django.conf import settings
from django.db import models


class APIChangeLog(models.Model):
    """
    Enables logs of API-based changes to be viewed in admin.
    Not meant to be used directly (see base.logging.py)
    Intentionally does not have UUID
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    username = models.CharField(max_length=200, blank=True)
    full_path = models.CharField(max_length=200, blank=True)
    method = models.CharField(max_length=30, blank=True)
    timestamp = models.DateTimeField(blank=True)
    data = models.JSONField(default=dict)
    params = models.JSONField(default=dict)

    def __str__(self):
        return f"{str(self.method)}: {str(self.full_path)}"

    class Meta:
        verbose_name = "API Change Log"


class AppErrorLog(models.Model):
    LOG_LEVELS = (
        ("WARNING", "Warning"),
        ("ERROR", "Error"),
        ("FATAL", "Fatal"),
    )

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    level = models.CharField(max_length=10, choices=LOG_LEVELS)
    message = models.CharField(max_length=200)
    stack_trace = models.TextField(
        blank=True, null=True, help_text="Stack trace error logging if it's available"
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        on_delete=models.DO_NOTHING,
        help_text="User data if available",
    )
    client_data = models.TextField(
        blank=True,
        null=True,
        help_text="Any other client data, user agent, ip address etc",
    )
    timestamp = models.DateTimeField(auto_now_add=True)

    class Meta:
        verbose_name = "App Error Log"



from datetime import datetime

from django.test import TestCase

from smoothglue.authentication.models import PlatformUser
from smoothglue.duplication.exceptions import DuplicationError
from smoothglue.duplication.utils import (
    duplicate_object,
    get_parent_field_name,
    setup_fields_to_duplicate,
)
from smoothglue.tracker.models import APIChangeLog, AppErrorLog


class TestDuplication(TestCase):
    def setUp(self):
        self.test_item = APIChangeLog.objects.create(
            username="tester1", data={"foo": "bar"}, timestamp=datetime.now()
        )

    def test_valid_duplication_with_map(self):
        new_test_item = duplicate_object(
            self.test_item,
            include_children=False,
            included_fields=["data", "timestamp"],
            field_mapper={"username": "tester45"},
        )
        new_test_item.save()

        self.assertEqual(len(APIChangeLog.objects.all()), 2)
        self.assertEqual(self.test_item.data, new_test_item.data)
        self.assertNotEqual(self.test_item.username, new_test_item.username)

    def test_valid_duplication_without_map(self):
        new_test_item = duplicate_object(
            self.test_item,
            include_children=False,
            included_fields=["data", "username", "timestamp"],
        )
        new_test_item.save()
        self.assertEqual(len(APIChangeLog.objects.all()), 2)

        self.assertEqual(self.test_item.data, new_test_item.data)
        self.assertEqual(self.test_item.username, new_test_item.username)

        self.assertNotEqual(self.test_item.id, new_test_item.id)

    def test_invalid_duplication_with_pk_field(self):
        with self.assertRaises(DuplicationError):
            duplicate_object(
                self.test_item,
                include_children=False,
                included_fields=["id", "data", "username", "timestamp"],
            )

        self.assertEqual(len(APIChangeLog.objects.all()), 1)

    def test_invalid_duplication_with_conflicting_fields(self):
        with self.assertRaises(DuplicationError):
            duplicate_object(
                self.test_item,
                include_children=False,
                included_fields=["data", "username", "timestamp"],
                field_mapper={"username": "tester42"},
            )
        self.assertEqual(len(APIChangeLog.objects.all()), 1)

    def test_invalid_duplication_with_nonexisting_fields(self):
        with self.assertRaises(DuplicationError):
            duplicate_object(
                self.test_item,
                include_children=False,
                included_fields=["data", "username", "foo", "timestamp"],
            )

            self.assertEqual(len(APIChangeLog.objects.all()), 1)

        with self.assertRaises(DuplicationError):
            duplicate_object(
                self.test_item,
                include_children=False,
                included_fields=["data", "username", "timestamp"],
                field_mapper={"foo": 0},
            )

            self.assertEqual(len(APIChangeLog.objects.all()), 1)

    def test_invalid_duplication_with_incorrect_field_types(self):
        with self.assertRaises(DuplicationError):
            duplicate_object(
                self.test_item,
                include_children=False,
                included_fields=["username", "data"],
                field_mapper={"timestamp": "1000"},
            )

        self.assertEqual(len(APIChangeLog.objects.all()), 1)


class TestDuplicationUtils(TestCase):
    def setUp(self):
        self.test_item = PlatformUser.objects.create(email="foo@test.com")

    def test_get_parent_field_name(self):

        # valid relationship
        parent_field_name = get_parent_field_name(self.test_item, AppErrorLog)
        self.assertEqual(parent_field_name, "user")

        # invalid relationship
        parent_field_name = get_parent_field_name(self.test_item, APIChangeLog)
        self.assertIsNone(parent_field_name)

    def test_setup_fields_to_duplicate_no_children(self):

        # no excluded fields
        included_fields, field_mapper = setup_fields_to_duplicate(
            self.test_item,
            False,
            included_fields=["data", "user_image"],
            excluded_fields=None,
            field_mapper={"email": "bar@test.com"},
        )

        # we expect no change
        self.assertEqual(included_fields, ["data", "user_image"])
        self.assertEqual(field_mapper, {"email": "bar@test.com"})

        # only excluded fields
        included_fields, field_mapper = setup_fields_to_duplicate(
            self.test_item,
            False,
            included_fields=None,
            excluded_fields=["data", "user_image"],
            field_mapper=None,
        )

        # we don't have the excluded fields but we do have others
        self.assertGreater(len(included_fields), 1)
        self.assertNotIn("data", included_fields)
        self.assertNotIn("user_image", included_fields)
        self.assertIn("email", included_fields)

        # no children added
        self.assertNotIn("platformorganization", included_fields)

        # field_mapper is made a dict still
        self.assertEqual(field_mapper, {})

    def test_setup_fields_to_duplicate_with_children(self):

        # no excluded fields
        included_fields, field_mapper = setup_fields_to_duplicate(
            self.test_item,
            True,
            included_fields=["data", "user_image"],
            excluded_fields=None,
            field_mapper={"email": "bar@test.com"},
        )

        # we expect no change
        self.assertEqual(included_fields, ["data", "user_image"])
        self.assertEqual(field_mapper, {"email": "bar@test.com"})

        # only excluded fields
        included_fields, field_mapper = setup_fields_to_duplicate(
            self.test_item,
            True,
            included_fields=None,
            excluded_fields=["data", "user_image"],
            field_mapper=None,
        )

        # we don't have the excluded fields but we do have others, including children
        self.assertGreater(len(included_fields), 1)
        self.assertNotIn("data", included_fields)
        self.assertNotIn("user_image", included_fields)
        self.assertIn("email", included_fields)
        self.assertIn("platformorganization", included_fields)

        # field_mapper is made a dict still
        self.assertEqual(field_mapper, {})

    def test_setup_fields_to_duplicate_include_or_exclude_children(self):
        included_fields_with_children, _ = setup_fields_to_duplicate(
            self.test_item,
            True,
            included_fields=None,
            excluded_fields=None,
            field_mapper=None,
        )

        included_fields_without_children, _ = setup_fields_to_duplicate(
            self.test_item,
            False,
            included_fields=None,
            excluded_fields=None,
            field_mapper=None,
        )

        self.assertTrue(
            set(included_fields_without_children).issubset(
                included_fields_with_children
            )
        )



# Generated by Django 4.2.19 on 2025-03-17 18:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0005_platformorganization_abbreviation_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='organizationcategory',
            name='data',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='platformorganization',
            name='data',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='platformuser',
            name='data',
            field=models.JSONField(blank=True, default=dict),
        ),
    ]



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from adrf.mixins import UpdateModelMixin as AsyncUpdateModelMixin
from asgiref.sync import sync_to_async
from rest_framework.mixins import DestroyModelMixin, UpdateModelMixin
from rest_framework.views import Request

from smoothglue.core.views import AsyncListModelViewSet, ListModelViewSet


@pytest.mark.parametrize(
    "sync_function_name, async_function_name",
    [
        ("list", "alist"),
        ("create", "acreate"),
        ("update", "aupdate"),
        ("partial_update", "partial_aupdate"),
        ("destroy", "adestroy"),
    ],
)
@pytest.mark.asyncio
@pytest.mark.django_db
@patch.object(DestroyModelMixin, "perform_destroy")
@patch.object(UpdateModelMixin, "perform_update")
@patch.object(AsyncUpdateModelMixin, "perform_aupdate", new_callable=AsyncMock)
async def test_sync_and_async_views_have_same_behavior(
    _mock_perform_aupdate,
    _mock_perform_update,
    _mock_perform_destroy,
    sync_function_name,
    async_function_name,
    get_test_model,
    get_test_sync_serializer,
    get_test_async_serializer,
):
    """
    Given ListModelViewSet and AsyncModelViewSet instances
    When we call equivalent functions (i.e. `list` and `alist`)
    Then we expecte the same results for both function calls
    And their behavior should match exactly

    NOTE: This is a smoke test for detecting ListModelViewSet and AsyncListModelViewSet
    diverging in behavior. They should match.
    """
    # Setup
    SomeModel = get_test_model
    SomeSyncSerializer = get_test_sync_serializer
    SomeAsyncSerializer = get_test_async_serializer
    SOME_NAME = "some name"
    SOME_OBJECT = {"pk": 1, "name": SOME_NAME}

    class SomeListModelViewSet(ListModelViewSet):
        serializer_class = SomeSyncSerializer

        def get_queryset(self):
            return None

    class SomeAsyncListModelViewSet(AsyncListModelViewSet):
        serializer_class = SomeAsyncSerializer

        def get_queryset(self):
            return None

    sync_view_set = SomeListModelViewSet()
    sync_view_set.request = None
    sync_view_set.format_kwarg = None
    sync_view_set.kwargs = SOME_OBJECT

    async_view_set = SomeAsyncListModelViewSet()
    async_view_set.request = None
    async_view_set.format_kwarg = None
    async_view_set.kwargs = SOME_OBJECT

    some_model_instance = SomeModel(pk=1, name=SOME_NAME)
    mock_request = MagicMock(spec=Request)
    mock_request.data = {"id": 1, "name": SOME_NAME}
    mock_request._request = MagicMock()

    # Test
    with patch(
        "rest_framework.generics.get_object_or_404", return_value=some_model_instance
    ):
        sync_result = await sync_to_async(getattr(sync_view_set, sync_function_name))(
            mock_request
        )

    match async_function_name:
        case "adestroy":  # adestroy calls sync function under the hood
            with patch(
                "rest_framework.generics.get_object_or_404",
                return_value=some_model_instance,
            ):
                async_result = await getattr(async_view_set, async_function_name)(
                    mock_request
                )
        case _:
            with patch(
                "adrf.generics.aget_object_or_404",
                new_callable=AsyncMock,
                return_value=some_model_instance,
            ):
                async_result = await getattr(async_view_set, async_function_name)(
                    mock_request
                )

    assert sync_result.status_code == async_result.status_code

    match sync_function_name:
        # For some reason `data` is sometimes returned as a function or property
        case "create" | "update" | "partial_update":
            assert sync_result.data() == async_result.data
        case _:
            assert sync_result.data == async_result.data



from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.db import models
from django.utils.html import format_html_join
from django.utils.safestring import mark_safe
from django_json_widget.widgets import JSONEditorWidget

from smoothglue.authentication.models import (
    OrganizationCategory,
    OrganizationMember,
    PlatformOrganization,
)
from smoothglue.core.admin import BaseAuditModelAdmin

User = get_user_model()


class OrganizationMembersInline(admin.TabularInline):
    model = OrganizationMember
    readonly_fields = BaseAuditModelAdmin.readonly_fields
    fk_name = "platform_user"


class UserAdmin(BaseUserAdmin):
    """
    Enhanced admin view for the User model that provides additional
    features like filtering, searching, and a custom layout for JSON fields.
    Some fields are read-only as they can only be changed in Keycloak.
    """

    readonly_fields = (
        "username",
        "email",
    )
    list_display = (
        "username",
        "email",
        "first_name",
        "last_name",
        "is_staff",
    )
    search_fields = (
        "username",
        "email",
        "first_name",
        "last_name",
    )
    list_filter = (
        "is_staff",
        "is_superuser",
        "is_active",
        "groups",
    )
    formfield_overrides = {
        models.JSONField: {"widget": JSONEditorWidget},
    }

    fieldsets = (
        (None, {"fields": ("username", "email", "password")}),
        ("Personal info", {"fields": (("first_name", "last_name"), "user_image")}),
        (
            "Permissions",
            {
                "fields": (
                    "is_active",
                    "is_staff",
                    "is_superuser",
                    "groups",
                    "user_permissions",
                )
            },
        ),
        ("Important dates", {"fields": ("last_login", "date_joined")}),
        ("Custom Data", {"fields": ("data",)}),
    )
    inlines = (OrganizationMembersInline,)


class PlatformOrganizationsInline(admin.TabularInline):
    model = PlatformOrganization
    readonly_fields = BaseAuditModelAdmin.readonly_fields + ("group",)


class PlatformOrganizationMembersInline(admin.TabularInline):
    model = OrganizationMember
    readonly_fields = BaseAuditModelAdmin.readonly_fields


class PlatformOrganizationAdmin(BaseAuditModelAdmin):
    model = PlatformOrganization
    inlines = (
        PlatformOrganizationsInline,
        PlatformOrganizationMembersInline,
    )
    list_display = (
        "name",
        "parent_organization",
        "org_category",
        "abbreviation",
        "child_organizations",
    )
    list_editable = ("parent_organization", "org_category", "abbreviation")

    def child_organizations(self, obj):
        return format_html_join(
            mark_safe("<br>"),
            "<a href='/admin/authentication/platformorganization/{}'>{}</>",
            (
                (
                    org.id,
                    org.name,
                )
                for org in obj.child_organizations.all()
            ),
        ) or mark_safe("-")

    readonly_fields = BaseAuditModelAdmin.readonly_fields + ("group",)
    classes = ["wide", "extrapretty"]


class OrganizationCategoryAdmin(BaseAuditModelAdmin):
    model = OrganizationCategory
    list_display = ("name", "description")


admin.site.register(User, UserAdmin)
admin.site.register(PlatformOrganization, PlatformOrganizationAdmin)
admin.site.register(OrganizationCategory, OrganizationCategoryAdmin)



import asyncio
from unittest.mock import MagicMock

import pytest
from django.contrib.auth import get_user_model
from rest_framework.request import Request

from smoothglue.core.serializers.abstract import (
    AsyncAuditSerializer,
    AsyncDefaultSerializer,
    AuditSerializer,
    DefaultSerializer,
)


@pytest.mark.parametrize(
    "sync_function_name, async_function_name, args, kwargs",
    [
        ("validate", "validate", [], {"attrs": {"name": "some_name"}}),
        ("to_representation", "ato_representation", [], {"instance": "get_instance"}),
        ("to_representation", "to_representation", [], {"instance": "get_instance"}),
    ],
)
@pytest.mark.asyncio
async def test_sync_and_async_serializers_have_same_behavior(
    sync_function_name, async_function_name, args, kwargs, get_test_model
):
    """
    Given an instance of a DefaultSerializer and AsyncDefaultSerializer
    When call any given overridden function on both instances
    Then we expect the output and behavior to be the same
    """
    SomeModel = get_test_model

    if kwargs.get("instance") == "get_instance":
        kwargs["instance"] = SomeModel(pk=1, name="some_name")

    class SomeSyncDefaultSerializer(DefaultSerializer):
        class Meta:
            model = SomeModel
            fields = ["name"]

    class SomeAsyncDefaultSerializer(AsyncDefaultSerializer):
        class Meta:
            model = SomeModel
            fields = ["name"]

    some_sync_serializer = SomeSyncDefaultSerializer()
    some_async_serializer = SomeAsyncDefaultSerializer()

    sync_result = getattr(some_sync_serializer, sync_function_name)(*args, **kwargs)

    async_function = getattr(some_async_serializer, async_function_name)
    if asyncio.iscoroutinefunction(async_function):
        async_result = await async_function(*args, **kwargs)
    else:
        async_result = async_function(*args, **kwargs)

    assert sync_result == async_result


@pytest.mark.parametrize(
    "sync_function_name, async_function_name, args, kwargs",
    [
        ("create", "acreate", [], {"validated_data": {"name": "some_name"}}),
        (
            "update",
            "aupdate",
            [],
            {"instance": "get_instance", "validated_data": {"name": "some_name"}},
        ),
    ],
)
@pytest.mark.asyncio
async def test_sync_and_async_audit_serializers_have_same_behavior(
    sync_function_name,
    async_function_name,
    args,
    kwargs,
    get_test_audit_model,
    patch_db_operations,
):
    """
    Given an instance of an AuditSerializer and an AsyncAuditSerializer
    When we call the overridden functions on the serializer instances with the same inputs
    Then we expect the same outputs and behavior
    """
    # Setup
    SomeAuditModel = get_test_audit_model
    some_audit_model_instance = SomeAuditModel(name="some_name")

    if kwargs.get("instance") == "get_instance":
        kwargs["instance"] = some_audit_model_instance

    UserModel = get_user_model()

    some_user = UserModel()

    class SomeSyncAuditSerializer(AuditSerializer):
        class Meta:
            model = SomeAuditModel
            fields = AuditSerializer.Meta.fields + ["name"]

    class SomeAsyncAuditSerializer(AsyncAuditSerializer):
        class Meta:
            model = SomeAuditModel
            fields = AsyncAuditSerializer.Meta.fields + ["name"]

    mock_request = MagicMock(spec=Request)
    mock_request.user = some_user

    context = {"request": mock_request}
    some_sync_serializer = SomeSyncAuditSerializer(context=context)

    some_async_serializer = SomeAsyncAuditSerializer(context=context)

    # Test
    with patch_db_operations(return_value=some_audit_model_instance):
        sync_result = getattr(some_sync_serializer, sync_function_name)(*args, **kwargs)

        async_result = await getattr(some_async_serializer, async_function_name)(
            *args, **kwargs
        )

    assert sync_result == async_result



# tests/test_admin.py (or your preferred test file location)

import unittest
from unittest.mock import Mock

# Assuming your admin.py is in 'smoothglue.tracker.admin'
# and models are in 'smoothglue.tracker.models'
# Adjust these imports based on your actual project structure.
from smoothglue.tracker.admin import APIChangeLogAdmin, AppErrorLogAdmin
from smoothglue.tracker.models import APIChangeLog, AppErrorLog

# Mock the admin site
mock_admin_site = Mock()


class APIChangeLogAdminTests(unittest.TestCase):
    def setUp(self):
        self.admin_instance = APIChangeLogAdmin(
            model=APIChangeLog, admin_site=mock_admin_site
        )
        self.mock_request_superuser = Mock()
        self.mock_request_superuser.user = Mock()
        self.mock_request_superuser.user.is_superuser = True

        self.mock_request_non_superuser = Mock()
        self.mock_request_non_superuser.user = Mock()
        self.mock_request_non_superuser.user.is_superuser = False

        self.mock_object = Mock(spec=APIChangeLog)  # A mock model instance

    def test_attributes_set_correctly(self):
        self.assertEqual(self.admin_instance.search_fields, ["full_path"])
        self.assertEqual(
            self.admin_instance.list_display,
            ("full_path", "username", "method", "timestamp", "data"),
        )
        self.assertEqual(
            self.admin_instance.list_filter, ("username", "method", "timestamp")
        )
        self.assertEqual(
            self.admin_instance.readonly_fields,
            ["username", "full_path", "method", "timestamp", "data", "params"],
        )
        self.assertEqual(self.admin_instance.ordering, ("-timestamp",))

    def test_has_add_permission(self):
        self.assertFalse(
            self.admin_instance.has_add_permission(self.mock_request_superuser)
        )
        self.assertFalse(
            self.admin_instance.has_add_permission(self.mock_request_non_superuser)
        )

    def test_has_change_permission(self):
        self.assertFalse(
            self.admin_instance.has_change_permission(self.mock_request_superuser)
        )
        self.assertFalse(
            self.admin_instance.has_change_permission(
                self.mock_request_superuser, obj=self.mock_object
            )
        )
        self.assertFalse(
            self.admin_instance.has_change_permission(self.mock_request_non_superuser)
        )
        self.assertFalse(
            self.admin_instance.has_change_permission(
                self.mock_request_non_superuser, obj=self.mock_object
            )
        )

    def test_has_view_permission_superuser(self):
        self.assertTrue(
            self.admin_instance.has_view_permission(self.mock_request_superuser)
        )
        self.assertTrue(
            self.admin_instance.has_view_permission(
                self.mock_request_superuser, obj=self.mock_object
            )
        )

    def test_has_view_permission_non_superuser(self):
        self.assertFalse(
            self.admin_instance.has_view_permission(self.mock_request_non_superuser)
        )
        self.assertFalse(
            self.admin_instance.has_view_permission(
                self.mock_request_non_superuser, obj=self.mock_object
            )
        )


class AppErrorLogAdminTests(unittest.TestCase):
    def setUp(self):
        self.admin_instance = AppErrorLogAdmin(
            model=AppErrorLog, admin_site=mock_admin_site
        )
        self.mock_request_superuser = Mock()
        self.mock_request_superuser.user = Mock()
        self.mock_request_superuser.user.is_superuser = True

        self.mock_request_non_superuser = Mock()
        self.mock_request_non_superuser.user = Mock()
        self.mock_request_non_superuser.user.is_superuser = False

        self.mock_object = Mock(spec=AppErrorLog)  # A mock model instance

    def test_attributes_set_correctly(self):
        self.assertEqual(self.admin_instance.search_fields, ["user"])
        self.assertEqual(
            self.admin_instance.list_display, ("level", "message", "user", "timestamp")
        )
        self.assertEqual(
            self.admin_instance.list_filter, ("user", "level", "timestamp")
        )
        self.assertEqual(
            self.admin_instance.readonly_fields,
            ["user", "level", "message", "stack_trace", "client_data", "timestamp"],
        )
        self.assertEqual(self.admin_instance.ordering, ("-timestamp",))

    def test_has_add_permission(self):
        self.assertFalse(
            self.admin_instance.has_add_permission(self.mock_request_superuser)
        )
        self.assertFalse(
            self.admin_instance.has_add_permission(self.mock_request_non_superuser)
        )

    def test_has_change_permission(self):
        self.assertFalse(
            self.admin_instance.has_change_permission(self.mock_request_superuser)
        )
        self.assertFalse(
            self.admin_instance.has_change_permission(
                self.mock_request_superuser, obj=self.mock_object
            )
        )
        self.assertFalse(
            self.admin_instance.has_change_permission(self.mock_request_non_superuser)
        )
        self.assertFalse(
            self.admin_instance.has_change_permission(
                self.mock_request_non_superuser, obj=self.mock_object
            )
        )

    def test_has_view_permission_superuser(self):
        self.assertTrue(
            self.admin_instance.has_view_permission(self.mock_request_superuser)
        )
        self.assertTrue(
            self.admin_instance.has_view_permission(
                self.mock_request_superuser, obj=self.mock_object
            )
        )

    def test_has_view_permission_non_superuser(self):
        self.assertFalse(
            self.admin_instance.has_view_permission(self.mock_request_non_superuser)
        )
        self.assertFalse(
            self.admin_instance.has_view_permission(
                self.mock_request_non_superuser, obj=self.mock_object
            )
        )



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from adrf.serializers import BaseSerializer
from django.db.models.manager import Manager
from rest_framework.request import Request

from smoothglue.core.serializers.abstract import (
    AsyncAuditSerializer,
    AsyncDefaultSerializer,
)
from smoothglue.core.views import AsyncListModelViewSet


@pytest.mark.parametrize(
    "serializer_class", [AsyncDefaultSerializer, AsyncAuditSerializer]
)
@pytest.mark.parametrize(
    "calling_function_name, patch_target, new_callable",
    [
        (
            "acreate",
            "adrf.serializers.ModelSerializer.validate",
            None,
        ),
        (
            "acreate",
            "adrf.serializers.ModelSerializer.validate",
            None,
        ),
        (
            "acreate",
            "adrf.serializers.ModelSerializer.ato_representation",
            AsyncMock,
        ),
        (
            "alist",
            "adrf.serializers.ModelSerializer.ato_representation",
            AsyncMock,
        ),
        (
            "aupdate",
            "adrf.serializers.ModelSerializer.validate",
            None,
        ),
        (
            "aupdate",
            "adrf.serializers.ModelSerializer.ato_representation",
            AsyncMock,
        ),
        (
            "partial_aupdate",
            "adrf.serializers.ModelSerializer.validate",
            None,
        ),
        (
            "partial_aupdate",
            "adrf.serializers.ModelSerializer.ato_representation",
            AsyncMock,
        ),
    ],
)
@pytest.mark.asyncio
@pytest.mark.django_db
async def test_async_serializer_functions_called(
    serializer_class, calling_function_name, patch_target, new_callable, get_test_model
):
    """
    Given a serializer that is used by an async view
    When we can call any of the overridden functions
    Then we expect the top level function to also be called
    """
    # Setup
    SomeModel = get_test_model

    class AsyncSomeSerializer(serializer_class):
        class Meta:
            model = SomeModel
            fields = ["name"]

    class AsyncSomeModelViewSet(AsyncListModelViewSet):
        serializer_class = AsyncSomeSerializer

        def get_queryset(self):
            return {SomeModel(pk=1, name="some_name")}

    mock_request = MagicMock(spec=Request)
    mock_request.data = {"pk": 1, "name": "some_name"}
    mock_request._request = MagicMock()
    mock_request._request.path = "path"
    mock_request._request.method = "method"
    mock_request.query_params.dict.return_value = {}

    async_view = AsyncSomeModelViewSet()
    async_view.request = mock_request
    async_view.format_kwarg = None
    async_view.kwargs = {"pk": 1}

    some_model_instance = SomeModel(pk=1, name="some_name")
    # Mocking away access to database layer
    with (
        patch(
            "adrf.generics.aget_object_or_404",
            new_callable=AsyncMock,
            return_value=some_model_instance,
        ),
        patch(
            "rest_framework.generics.get_object_or_404",
            return_value=some_model_instance,
        ),
        patch.object(
            BaseSerializer,
            "acreate",
            new_callable=AsyncMock,
            return_value=some_model_instance,
        ),
        patch.object(
            Manager,
            "create",
            return_value=some_model_instance,
        ),
    ):
        # Test
        with patch(
            patch_target, new_callable=new_callable, return_value={"name": "some_name"}
        ) as mock_func:
            assert async_view.view_is_async
            await getattr(async_view, calling_function_name)(mock_request, partial=True)
            mock_func.assert_called()



from html import unescape
from typing import Any

from adrf.serializers import ModelSerializer
from django.db.models import Model
from django.utils.html import escape

from smoothglue.core.models import AuditModel, TimeAuditModel


class AsyncDefaultSerializer(ModelSerializer):
    """
    A serializer that escapes HTML in input data and unescapes HTML in output data.
    This serializer is designed to be used with asynchronous views.
    """

    def validate(self, attrs: dict[str, Any]) -> dict[str, Any]:
        """
        Validate and escape HTML in the input data.

        Args:
            attrs (dict[str, Any]): The input data to validate.

        Returns:
            dict[str, Any]: The validated and escaped data.
        """
        for key, value in attrs.items():
            if isinstance(value, str):
                attrs[key] = escape(value)
        return super().validate(attrs)

    async def ato_representation(self, instance: Model) -> str:
        """
        Asynchronously convert the instance to its representation
        and unescape HTML in the output data.

        Args:
            instance (Model): The instance to convert.

        Returns:
            str: The unescaped representation of the instance.
        """
        data = await super().ato_representation(instance)
        for key, value in data.items():
            if isinstance(value, str):
                data[key] = unescape(value)
        return data

    def to_representation(self, instance: Model) -> str:
        """
        Convert the instance to its representation and unescape HTML in the output data.

        Args:
            instance (Model): The instance to convert.

        Returns:
            str: The unescaped representation of the instance.
        """
        data = super().to_representation(instance)
        for key, value in data.items():
            if isinstance(value, str):
                data[key] = unescape(value)
        return data

    class Meta:
        """
        Meta options for the serializer.
        """

        abstract = True


class AsyncTimeAuditSerializer(AsyncDefaultSerializer):
    """
    Serializer for the TimeAuditModel.
    Serializes the 'created_at' and 'updated_at' fields.
    """

    class Meta:
        """
        Meta options for the TimeAuditSerializer.
        """

        model = TimeAuditModel
        fields = ["created_at", "updated_at"]
        read_only_fields = fields
        abstract = True


class AsyncAuditSerializer(AsyncDefaultSerializer):
    """
    Serializer for the AuditModel.
    Serializes the fields from TimeAuditSerializer and adds 'created_by' and 'updated_by' fields.
    """

    class Meta:
        """
        Meta options for the AuditSerializer.
        """

        model = AuditModel
        fields = AsyncTimeAuditSerializer.Meta.fields + ["created_by", "updated_by"]
        read_only_fields = fields
        abstract = True

    async def acreate(self, validated_data: dict[str, Any]) -> Model:
        """
        Asynchronously create a new AuditModel instance,
        setting the 'created_by' and 'updated_by' fields.

        Args:
            validated_data (dict[str, Any]): The validated data to create the instance with.

        Returns:
            Model: The newly created AuditModel instance.
        """
        user = self.context["request"].user
        if user:
            validated_data["updated_by"] = user
            validated_data["created_by"] = user
        return await super().acreate(validated_data)

    async def aupdate(self, instance: Model, validated_data: dict[str, Any]) -> Model:
        """
        Asynchronously update an existing AuditModel instance, setting the 'updated_by' field.

        Args:
            instance (Model): The existing AuditModel instance.
            validated_data (dict[str, Any]): The validated data to update the instance with.

        Returns:
            Model: The updated AuditModel instance.
        """
        user = self.context["request"].user
        if user:
            instance.updated_by = user
        return await super().aupdate(instance, validated_data)



"""
This file sets up and configures Django.
It's used by scripts that need to execute as if running in a Django server.
"""

import django
from django.conf import settings

from smoothglue.core.scripts import test_settings


def boot_django():
    # The settings.configure() call takes a list of args that are equivalent
    # to variables defined in a test_settings.py file.
    # Anything you would need in your test_settings.py to make your app run gets passed here
    settings.configure(
        BASE_DIR=test_settings.BASE_DIR,
        DEBUG=test_settings.DEBUG,
        DATABASES=test_settings.DATABASES,
        INSTALLED_APPS=test_settings.INSTALLED_APPS,
        TIME_ZONE=test_settings.TIME_ZONE,
        USE_TZ=test_settings.USE_TZ,
        ROOT_URLCONF=test_settings.ROOT_URLCONF,
        AUTH_USER_MODEL=test_settings.AUTH_USER_MODEL,
        SECRET_KEY=test_settings.SECRET_KEY,
        REMOTE_USER_HEADER=test_settings.REMOTE_USER_HEADER,
    )
    django.setup()



from django.contrib import admin
from django.db import models
from django_json_widget.widgets import JSONEditorWidget


class DataAdmin(admin.ModelAdmin):
    """
    This provides an easier WYSIWYG JSON Editor for JSONField
    """

    formfield_overrides = {
        models.JSONField: {"widget": JSONEditorWidget},
    }


class AuditTabularInline(admin.TabularInline):
    """
    A custom TabularInline that automatically makes 'created_by' and 'updated_by' as readonly fields.
    """

    readonly_fields = ("updated_by", "created_by")


class BaseAuditModelAdmin(DataAdmin):
    """
    This base admin class will disable the audit fields and will auto populate
    the audit fields based on the request users.
    """

    readonly_fields = ("created_by", "updated_by", "created_at", "updated_at")

    def save_model(self, request, obj, form, change):
        if not change:
            obj.created_by = request.user
        obj.updated_by = request.user
        return super().save_model(request, obj, form, change)

    def save_formset(self, request, form, formset, change):
        instances = formset.save(commit=False)
        for instance in instances:
            if not hasattr(instance, "created_by"):
                instance.created_by = request.user
            instance.updated_by = request.user
            instance.save()

        # Delete inlines objects
        for form_obj in formset.deleted_forms:
            if form_obj.instance.pk:
                form_obj.instance.delete()

        # Update inlines objects to have the updated_by
        for form_obj in formset.changed_objects:
            instance = form_obj[0]
            if not hasattr(instance, "updated_by"):
                instance.updated_by = request.user
                instance.save()

        # Update inlines objects to have the created_by and updated_by
        for form_obj in formset.new_objects:
            instance = form_obj
            if not hasattr(instance, "created_by"):
                instance.created_by = request.user
                instance.updated_by = request.user
                instance.save()

        super().save_formset(request, form, formset, change)



from ._abstract import AuditSerializer, DefaultSerializer, TimeAuditSerializer
from ._async_abstract import AsyncAuditSerializer, AsyncDefaultSerializer

__all__ = [
    "DefaultSerializer",
    "TimeAuditSerializer",
    "AuditSerializer",
    "AsyncDefaultSerializer",
    "AsyncAuditSerializer",
]



import copy
import json
import types
from collections.abc import MutableMapping
from typing import Any, Dict, List, Type, TypeVar
from uuid import UUID

from django.core import serializers
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Model
from django.test import Client, TestCase
from rest_framework.serializers import BaseSerializer
from rest_framework.test import APIRequestFactory, force_authenticate
from rest_framework.viewsets import ModelViewSet

T = TypeVar("T")


def convert_dict_value_types(dictionary, value_type, convert_type):
    # Convert the type of any value in a dictionary
    for key, val in dictionary.items():
        if isinstance(val, value_type):
            dictionary[key] = convert_type(val)
    return dictionary


class CommonSerializer:
    def __init__(self, serializer: Type[BaseSerializer], context: dict = None):
        self.serializer = serializer
        self.test_case = TestCase()
        self.context = context

    def test_model_serializer_valid(
        self,
        model: Type[Model],
        required_fields: List[str],
        data: Dict[str, T],
    ):
        self.test_required_fields(required_fields, data)
        ser = self.serializer(data=data)
        if self.context:
            ser.context.update(self.context)
        self.test_case.assertTrue(
            ser.is_valid(), msg=f"Serializer data not valid for {data}\n{ser.errors}"
        )
        instance = ser.save()
        model_qs = model.objects.filter(id=instance.id)
        self.test_case.assertEqual(len(model_qs), 1)
        self.test_case.assertEqual(model_qs.all()[0].id, instance.id)

    def test_required_fields(self, required_fields: List[str], data: Dict[str, T]):
        for field in required_fields:
            data_copy = copy.deepcopy(data)
            del data_copy[field]
            ser = self.serializer(data=data_copy)

            self.test_case.assertFalse(
                expr=ser.is_valid(),
                msg=f"Required field not checked: {field}\n{ser.errors}",
            )


class CommonAPI:
    def __init__(self, client: Client):
        self.client = client
        self.test_case = TestCase()

    # pylint: disable=R0913
    def test_get_list(
        self,
        object_data: List[Dict[str, T]],
        model: Type[Model],
        endpoint: str,
        update_fields: Dict[str, Any] = None,
        ignore_object_keys: list[str] = None,
    ):
        created_objects = [model.objects.create(**d) for d in object_data]

        response = self.client.get(endpoint, format="json")
        self.test_case.assertEqual(response.status_code, 200)
        if ignore_object_keys:
            response_data = [
                self._delete_keys_from_dict(item, ignore_object_keys)
                for item in response.json()
            ]
        else:
            response_data = response.json()
        self.test_case.assertEqual(
            len(response_data),
            len(object_data),
            msg="Expected number of objects differs from number in response.",
        )

        # Get ids of objects and from response to compare
        response_ids = {d.pop("id") for d in response_data}
        obj_ids = {str(obj.id) for obj in created_objects}
        self.test_case.assertEqual(response_ids, obj_ids)

        for obj_dict in object_data:
            if update_fields:
                obj_dict.update(update_fields)
            # NOTE: Object IDs removed from response before comparison
            self.test_case.assertTrue(
                obj_dict in response_data,
                msg=f"\nObject:\n{obj_dict}\nnot found in response:\n{response_data}",
            )

    # pylint: disable=R0913
    def test_get_filter_list(
        self,
        object_data: List[Dict[str, T]],
        model: Type[Model],
        endpoint: str,
        filter_name: str,
        filter_value: str,
        assertion_override_func: types.FunctionType = None,
    ):
        for d in object_data:
            model.objects.create(**d)

        response = self.client.get(endpoint, format="json")

        filtered_endpoint = f"{endpoint}?{filter_name}={filter_value}"
        filtered_response = self.client.get(filtered_endpoint, format="json")
        self.test_case.assertEqual(response.status_code, 200)
        self.test_case.assertEqual(filtered_response.status_code, 200)

        # Get properties based on filter name and check for equality
        if assertion_override_func is None:
            self.test_case.assertNotEqual(
                len(response.json()),
                len(filtered_response.json()),
                msg="Expected number of objects are the same in filtered response.",
            )

            response_values = {d.pop(filter_name) for d in filtered_response.json()}
            for value in response_values:
                self.test_case.assertEqual(
                    value,
                    filter_value,
                    msg=f"\nFilter error: {value} is not the filtered value {filter_value}",
                )
        else:
            # This is mainly for adding support for asserting nested filters, which the
            # nested filter value are sometimes not included in the response.
            assertion_override_func(filtered_response, self.test_case)

    def test_post_create(
        self,
        object_data: Dict[str, T],
        endpoint: str,
        update_fields: Dict[str, Any] = None,
        ignore_object_keys: list[str] = None,
    ):
        response = self.client.post(endpoint, data={}, format="json")
        self.test_case.assertEqual(response.status_code, 400)

        response = self.client.post(endpoint, data=object_data, format="json")
        self.test_case.assertEqual(
            response.status_code, 201, msg=(f"Response Data:\n{response.data}")
        )

        if update_fields:
            object_data.update(update_fields)

        response.json().pop("id")
        self.test_case.assertEqual(
            object_data,
            self._delete_keys_from_dict(response.json(), ignore_object_keys),
        )

    def test_post_create_nested(
        self,
        object_data: Dict[str, Any],
        update_data: Dict[str, List[Dict[str, Any]]],
        endpoint: str,
    ):
        """Test the creation of nested objects in a single POST"""
        for field in update_data:
            for i, data in enumerate(update_data[field]):  # Get nested object
                for nested_field in data:
                    object_copy = copy.deepcopy(object_data)
                    update_copy = copy.deepcopy(update_data)
                    data_copy = data.copy()
                    del data_copy[nested_field]
                    update_copy[field][i] = data_copy
                    object_copy[field] = update_copy[field]

                    response = self.client.post(
                        endpoint, data=object_copy, format="json"
                    )
                    self.test_case.assertEqual(
                        response.status_code,
                        400,
                        msg=(
                            f"Creation of partial nested objects should not be valid "
                            f"(missing '{nested_field}'):\n{response.data}"
                        ),
                    )

        object_data.update(**update_data)

        response = self.client.post(endpoint, data=object_data, format="json")
        self.test_case.assertEqual(
            response.status_code,
            201,
            msg=(
                "Nested objects for creation should be valid.\n"
                f"Response Data:\n{response.data}"
            ),
        )

    def test_post_create_list(
        self,
        object_data: List[Dict[str, T]],
        endpoint: str,
        update_fields: Dict[str, Any] = None,
    ):
        response = self.client.post(endpoint, data=object_data, format="json")
        self.test_case.assertEqual(response.status_code, 201)
        self.test_case.assertEqual(len(response.json()), len(object_data))

        for response_obj in response.json():
            del response_obj["id"]

        for obj in object_data:
            if update_fields:
                obj.update(update_fields)
            self.test_case.assertTrue(
                obj in response.json(),
                f"Could not find\n{obj}\nin response:\n{response.json()}",
            )

    def test_get_by_id(
        self,
        object_data: Dict[str, T],
        model: Type[Model],
        endpoint: str,
        update_fields: Dict[str, Any] = None,
    ):
        # Create the object in db first
        created_object = model.objects.create(**object_data)
        response = self.client.get(f"{endpoint}{created_object.id}/", format="json")

        self.test_case.assertEqual(response.status_code, 200)

        del response.json()["id"]
        if update_fields:
            object_data.update(update_fields)
        self.test_case.assertEqual(object_data, response.json())

    def test_post_empty_nested_objects(
        self,
        object_data: Dict[str, T],
        nested_field_names: List[str],
        endpoint: str,
    ):
        for field_name in nested_field_names:
            object_copy = copy.deepcopy(object_data)
            object_copy.update(**{field_name: [{}]})
            response = self.client.post(endpoint, data=object_copy, format="json")
            self.test_case.assertEqual(
                response.status_code,
                400,
                msg="Empty nested objects should not be valid",
            )

    def test_put(
        self,
        original_object: T,
        new_data: Dict[str, T],
        endpoint: str,
        update_fields: Dict[str, Any] = None,
        ignore_object_keys: list[str] = None,
    ):
        response = self.client.put(endpoint, data=new_data, format="json")
        self.test_case.assertEqual(response.status_code, 200)

        # Test that we are doing PUT against the same object while also deleting ID from response
        self.test_case.assertEqual(str(original_object.id), response.json().pop("id"))

        if update_fields:
            new_data.update(update_fields)
        self.test_case.assertEqual(
            new_data, self._delete_keys_from_dict(response.json(), ignore_object_keys)
        )

    @staticmethod
    def _clean_up_nested_create(
        response, field, original_id, original_ids, update_data
    ):
        """
        Replaces ONE deleted object with a created one (from nested PUT replace)
        Updates 'update_data' in place
        """
        complete = False
        for obj in response.data[field]:
            if UUID(obj["id"]) not in original_ids:
                for j, update_obj in enumerate(update_data[field]):
                    if update_obj["id"] == original_id:
                        update_data[field][j] = obj
                        complete = True
                        break
            if complete:
                break

    @staticmethod
    def _get_original_ids(data):
        """
        Grabs all ids in a list of objects, converting them to UUID if necessary
        """
        return [
            obj["id"] if isinstance(obj["id"], UUID) else UUID(obj["id"])
            for obj in data
        ]

    @staticmethod
    def _delete_keys_from_dict(dictionary, keys):
        if keys is None:
            return dictionary

        keys_set = set(keys)  # Just an optimization for the "if key in keys" lookup.

        modified_dict = {}
        for key, value in dictionary.items():
            if key not in keys_set:
                if isinstance(value, MutableMapping):
                    modified_dict[key] = CommonAPI._delete_keys_from_dict(
                        value, keys_set
                    )
                else:
                    modified_dict[key] = (
                        value  # or copy.deepcopy(value) if a copy is desired for non-dicts.
                    )
        return modified_dict

    def test_put_nested(
        self,
        object_data: Dict[str, Any],
        update_data: Dict[str, List[Dict[str, Any]]],
        endpoint: str,
    ):
        """
        Test update to nested objects in a single PUT

        Expects a list of nested objects with ALL required fields
        """
        for field in update_data:
            for i, data in enumerate(update_data[field]):  # Get nested object
                for nested_field in data:
                    object_copy = copy.deepcopy(object_data)
                    update_copy = copy.deepcopy(update_data)
                    data_copy = data.copy()
                    del data_copy[nested_field]  # Delete a required field
                    update_copy[field][
                        i
                    ] = data_copy  # Replace the nested object to test
                    object_copy.update(update_copy)

                    response = self.client.put(
                        endpoint, data=object_copy, format="json"
                    )
                    if nested_field == "id":
                        # Get all original ids for cleanup later
                        original_id = data["id"]
                        original_ids = self._get_original_ids(update_data[field])
                        self.test_case.assertEqual(
                            response.status_code,
                            200,
                            msg=(
                                f"Nested creation (without 'id' field) should be valid on PUT "
                                f":\n{response.data}"
                            ),
                        )
                        # Cleanup data by replacing update data with created obj
                        self._clean_up_nested_create(
                            response, field, original_id, original_ids, update_data
                        )
                    else:
                        self.test_case.assertEqual(
                            response.status_code,
                            400,
                            msg=(
                                f"Partial nested updates should not be valid on PUT "
                                f"(missing '{nested_field}'):\n{response.data}"
                            ),
                        )

        object_data.update(**update_data)

        response = self.client.put(endpoint, data=object_data, format="json")
        self.test_case.assertEqual(
            response.status_code,
            200,
            msg=f"Nested objects for full update should be valid:\n{response.data}",
        )

    def test_patch(
        self, new_data: Dict[str, T], expected_data: Dict[str, T], endpoint: str
    ):
        response = self.client.patch(endpoint, data=new_data, format="json")
        self.test_case.assertEqual(
            response.status_code,
            200,
            msg=("Expected status 200.\n" f"Response Data:\n{response.data}"),
        )
        self.test_case.assertEqual(expected_data, response.json())

    def test_patch_nested(
        self,
        object_data: Dict[str, Any],
        update_data: Dict[str, List[Dict[str, Any]]],
        endpoint: str,
    ):
        """
        Test partial update to nested objects in single PATCH

        Expects a list of nested objects with ALL required fields
        """
        for field in update_data:
            for i, data in enumerate(update_data[field]):  # Get nested object
                for nested_field in data:
                    object_copy = copy.deepcopy(object_data)
                    update_copy = copy.deepcopy(update_data)
                    data_copy = data.copy()
                    del data_copy[nested_field]
                    update_copy[field][i] = data_copy
                    object_copy[field] = update_copy[field]

                    response = self.client.patch(
                        endpoint, data=object_copy, format="json"
                    )
                    if nested_field == "id":
                        self.test_case.assertEqual(
                            response.status_code,
                            400,
                            msg=(
                                f"'id' required for partial update to nested objects:"
                                f"\n{response.data}"
                            ),
                        )
                    else:
                        self.test_case.assertEqual(
                            response.status_code,
                            200,
                            msg=(
                                f"Partial updates to nested objects should be valid "
                                f"(missing '{nested_field}'):\n{response.data}"
                            ),
                        )

        object_data.update(**update_data)
        response = self.client.patch(endpoint, data=object_data, format="json")
        self.test_case.assertEqual(response.status_code, 200, response.data)

    def test_delete(self, original_object: T, model: Type[Model], endpoint: str):
        response = self.client.delete(endpoint, format="json")
        self.test_case.assertEqual(response.status_code, 204)
        with self.test_case.assertRaises(ObjectDoesNotExist):
            model.objects.get(id=original_object.id)


class CommonView:
    def __init__(self, user: T, view_set: Type[ModelViewSet]):
        self.standard_config = {
            "get": "list",
            "post": "create",
            "get_id": "retrieve",
            "put": "update",
            "patch": "partial_update",
            "delete": "destroy",
        }
        self.test_case = TestCase()
        self.factory = APIRequestFactory()
        self.content_type = "application/json"
        self.user = user
        self.view = view_set.as_view(actions=self.standard_config)

    def test_get_list(self, objects: List[T], serializer: Type[BaseSerializer]):
        request = self.factory.get("")
        force_authenticate(request, user=self.user)
        response = self.view(request)
        self.test_case.assertEqual(response.status_code, 200)

        for obj in objects:
            self.test_case.assertTrue(
                serializer(obj).data in response.data,
                msg=(
                    f"TeamLead object data should be in response.\n"
                    f"Object:\n{serializer(obj).data}\nResponse:\n{response.data}"
                ),
            )

    def test_get_id(self, obj: T, serializer: Type[BaseSerializer]):
        request = self.factory.get(f"/{obj.id}/")
        force_authenticate(request, user=self.user)
        response = self.view(request)
        self.test_case.assertEqual(response.status_code, 200)
        self.test_case.assertEqual(serializer(obj).data, response.data[0])

    def test_post(
        self,
        obj_data: Dict[str, T],
        non_required_fields: Dict[str, T] = None,
    ):
        request = self.factory.post(
            "", data=json.dumps(obj_data), content_type=self.content_type
        )
        force_authenticate(request, user=self.user)
        response = self.view(request)
        self.test_case.assertEqual(response.status_code, 201)

        if non_required_fields:
            obj_data.update(non_required_fields)
        del response.data["id"]
        response.data = convert_dict_value_types(response.data, UUID, str)
        self.test_case.assertEqual(obj_data, response.data)

    def test_post_list(
        self,
        objects_data: List[Dict[str, T]],
        non_required_fields: Dict[str, T] = None,
    ):
        request = self.factory.post(
            "", data=json.dumps(objects_data), content_type=self.content_type
        )
        force_authenticate(request, user=self.user)
        response = self.view(request)
        self.test_case.assertEqual(response.status_code, 201)

        for i, data in enumerate(response.data):
            del data["id"]
            response.data[i] = convert_dict_value_types(data, UUID, str)

        for obj_data in objects_data:
            obj_data.update(non_required_fields)
            self.test_case.assertTrue(obj_data in response.data)

    def test_put(
        self,
        obj: T,
        update_data: Dict[str, T],
        model: Type[Model],
        serializer: Type[BaseSerializer],
    ):
        request = self.factory.put(
            f"/{obj.id}/", data=json.dumps(update_data), content_type=self.content_type
        )
        force_authenticate(request, user=self.user)
        response = self.view(request, pk=obj.id)
        self.test_case.assertEqual(response.status_code, 200)

        obj = model.objects.get(id=obj.id)
        self.test_case.assertEqual(serializer(obj).data, response.data)

    def test_patch(
        self,
        obj: T,
        update_data: Dict[str, T],
        model: Type[Model],
        serializer: Type[BaseSerializer],
    ):
        request = self.factory.patch(
            f"/{obj.id}/", data=json.dumps(update_data), content_type="application/json"
        )
        force_authenticate(request, user=self.user)
        response = self.view(request, pk=obj.id)
        self.test_case.assertEqual(response.status_code, 200)

        team_lead = model.objects.get(id=obj.id)
        self.test_case.assertEqual(serializer(team_lead).data, response.data)

    def test_delete(self, obj: T):
        request = self.factory.delete(f"/{obj.id}/")
        force_authenticate(request, user=self.user)
        response = self.view(request, pk=obj.id)
        self.test_case.assertEqual(response.status_code, 204)


class CommonMigrator:
    def __init__(self):
        self.test_case = TestCase()

    @staticmethod
    def get_json(objects):
        return json.loads(serializers.serialize("json", objects))

    def _compare_old_new_data(self, old_data, new_data):
        """
        Expects list of serialized json objects to be compared 1:1
        with the exception of the 'model' field which should be different
        """
        self.test_case.assertEqual(len(old_data), len(new_data))
        for old_obj in old_data:
            found = False
            for new_obj in new_data:
                if new_obj["pk"] == old_obj["pk"]:
                    self.test_case.assertNotEqual(
                        old_obj.pop("model"), new_obj.pop("model")
                    )
                    self.test_case.assertEqual(old_obj, new_obj)
                    found = True
                    break
            self.test_case.assertTrue(
                found, msg=f"{old_obj} not found in new data transfer"
            )

    def test_objects_transferred(
        self, test_map, new_state, new_app_label, old_app_label
    ):
        # Check that all objects transferred
        for model_name, old_data in test_map.items():
            NewModel = new_state.apps.get_model(new_app_label, model_name)
            new_data = json.loads(serializers.serialize("json", NewModel.objects.all()))
            self._compare_old_new_data(old_data, new_data)

        # Check that old objects do not exist
        for model_name, _ in test_map.items():
            with self.test_case.assertRaises(
                LookupError, msg=f"{model_name} should no longer exist"
            ):
                new_state.apps.get_model(old_app_label, "model_name")

    def test_objects_do_not_exist(self, test_map, new_state, old_app_label):
        # Check that new objects do not exist after transfer
        for model_name, _ in test_map.items():
            with self.test_case.assertRaises(
                LookupError, msg=f"{model_name} should no longer exist"
            ):
                new_state.apps.get_model(old_app_label, "model_name")



import logging
from typing import Any, Dict

import jwt
from django.contrib.auth import get_user_model, validators
from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from django.http import HttpRequest
from drf_spectacular.extensions import OpenApiAuthenticationExtension
from drf_spectacular.plumbing import build_bearer_security_scheme_object
from rest_framework import authentication

import smoothglue.authentication.defaults as app_defaults
from smoothglue.core.utils import get_setting

logger = logging.getLogger(__name__)


def decode_jwt(request: HttpRequest) -> Dict[str, Any]:
    """
    Decodes the JWT token from the HTTP request's authorization header.
    This function expects the token to be sent either as a Bearer token
    in the 'Authorization' header or directly in the 'Jwt' header.
    Verification of the token is not performed as it is handled externally
    by Istio and Keycloak before the request reaches the docker container.

    Parameters:
    request (HttpRequest): The HTTP request object containing the JWT token.

    Returns:
    Dict[str, Any]: The decoded JWT token.

    Raises:
    ValueError: If no JWT token is found in the headers.
    """
    encoded_auth_header = request.headers.get("Authorization")
    encoded_jwt_header = request.headers.get("Jwt")
    bearer_prefix = "Bearer "

    if encoded_auth_header and encoded_auth_header.startswith(bearer_prefix):
        token = encoded_auth_header[len(bearer_prefix) :]
    elif encoded_jwt_header:
        token = encoded_jwt_header
    else:
        raise ValueError("No JWT provided in header.")

    try:
        auth_header = jwt.decode(token, options={"verify_signature": False})
    except jwt.DecodeError as e:
        # pylint: disable=W0707
        raise ValueError(f"Invalid JWT token: {e}")

    return auth_header


def get_unknown_user():
    """
    For local development, a valid JWT is not present when testing the
    backend system, therefore use an UnknownUser as default user. This
    also serves as a fallbcak if a JWT is not sent from Istio.

    This fallback will allow the system to return view only data but
    not allow the user to add/update/delete anything.

    NOTE: No type specified because Django user management magic means
    specifying the type as 'User' is bad practice. Embrace magic.
    """
    user = get_user_model().objects.get_or_create(
        username="UnknownUser",
        email=get_setting("SINGLE_USER_DEFAULT_EMAIL", app_defaults),
        is_superuser=True,
        is_staff=True,
        defaults={
            "first_name": "Unknown",
            "last_name": "User",
            "data": {
                "rank": "N/A",
                "usercertificate": "N/A",
                "affiliation": "N/A",
            },
        },
    )[0]
    return user


class PlatformAuthentication(authentication.BaseAuthentication):
    """
    All authorization for Platform comes through the Authorization
    header. Therefore, the need for Django model control is not necessary
    as we can tie a user with a JWT to a Django user.
    """

    def authenticate(self, request: HttpRequest) -> tuple[Any, bool] | None:
        if get_setting("ENABLE_SINGLE_USER_MODE", app_defaults):
            return (get_unknown_user(), None)

        jwt_header = request.headers.get("Authorization") or request.headers.get("Jwt")

        if jwt_header:
            jwt_decoded = decode_jwt(request)
            email = jwt_decoded.get("email")
            preferred_username = jwt_decoded.get(
                "preferred_username", email.split("@")[0]
            )

            if not self.is_valid_email(email) or not self.is_valid_username(
                preferred_username
            ):
                return None

            user, created = get_user_model().objects.update_or_create(
                username=preferred_username.replace(" ", "."),
                defaults={
                    "first_name": jwt_decoded.get("given_name", ""),
                    "last_name": jwt_decoded.get("family_name", ""),
                    "email": email,
                    "data": {
                        "rank": jwt_decoded.get("rank", "N/A"),
                        "usercertificate": jwt_decoded.get("usercertificate", "N/A"),
                        "affiliation": jwt_decoded.get("affiliation", "N/A"),
                    },
                },
            )

            return (user, created)

        return None

    def is_valid_email(self, email):
        """
        Checks if email passes our validation rules
        """
        try:
            validate_email(email)
            return True
        except ValidationError:
            logger.info("Email failed unicode validator", stack_info=True)
            return False

    def is_valid_username(self, username):
        """
        Checks if username passes our validation rules
        """
        try:
            username_validator = validators.UnicodeUsernameValidator()
            username_validator(username)
            return True
        except ValidationError:
            logger.info("Username failed unicode validator", stack_info=True)
            return False


class PlatformAuthenticationScheme(OpenApiAuthenticationExtension):
    """
    Custom class for handling authentication and authorization beyond OpenAPI specification.
    """

    target_class = "smoothglue.authentication.auth.PlatformAuthentication"
    name = "tokenAuth"
    match_subclasses = True
    priority = -1

    def get_security_definition(self, auto_schema):
        return build_bearer_security_scheme_object(
            header_name="Authorization",
            token_prefix="Bearer",
        )


class RootAdminBackend(ModelBackend):
    """
    Django authentication backend to allow setting a superuser username and password via settings.

    Allows a user to authenticate using the username `settings.ROOT_ADMIN_USERNAME`, and
    password `settings.ROOT_ADMIN_PASSWORD`. Useful for creating a bootstrap user in situations
    where setting an environment variable may be preferable to creating a static fixture,
    or having to manually run `manage.py createsuperuser`.
    """

    def authenticate(self, request, username=None, password=None, **kwargs):
        username_valid = username == get_setting("ROOT_ADMIN_USERNAME", app_defaults)
        admin_password = get_setting("ROOT_ADMIN_PASSWORD", app_defaults)
        pwd_valid = admin_password is not None and password == admin_password
        if username_valid and pwd_valid:
            user, _ = get_user_model().objects.get_or_create(username=username)
            user.email = "admin@localhost"
            user.is_staff = True
            user.is_superuser = True
            user.save()
            return user
        return None


def platform_authentication_middleware(get_response):
    """
    Middleware which acts as a wrapper for smoothglue.authentication.auth.PlatformAuthentication.

    smoothglue.authentication.auth.PlatformAuthentication is a DRF authentication backend to allow
    authentication via JWT. This middleware extends the functionality of PlatformAuthentication from
    just being a DRF authentication mechanism to be able to authenticate any view.
    """

    def middleware(request):
        authenticated_user = hasattr(request, "user") and request.user.is_authenticated
        # pylint: disable=protected-access
        cached_user = (
            hasattr(request, "_cached_user") and request._cached_user.is_authenticated
        )
        if not (authenticated_user or cached_user):
            jwt_user = PlatformAuthentication().authenticate(request)
            if jwt_user is not None:
                # pylint: disable=protected-access
                request._cached_user = jwt_user[0]
                request.user = jwt_user[0]
        return get_response(request)

    return middleware



# Generated by Django 4.2.13 on 2024-07-07 16:55

import uuid

import django.db.models.constraints
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("authentication", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="OrganizationMember",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(class)ss_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="PlatformOrganization",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(class)ss_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="auth.group",
                    ),
                ),
                (
                    "parent_organization",
                    models.ForeignKey(
                        blank=True,
                        help_text="Parent Organization",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="child_organizations",
                        to="authentication.platformorganization",
                    ),
                ),
                (
                    "platform_users",
                    models.ManyToManyField(
                        default=None,
                        through="authentication.OrganizationMember",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(class)ss_updated",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="organizationmember",
            name="platform_organization",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="platform_organizations",
                to="authentication.platformorganization",
            ),
        ),
        migrations.AddField(
            model_name="organizationmember",
            name="platform_user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="platform_users",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="organizationmember",
            name="updated_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(class)ss_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddConstraint(
            model_name="organizationmember",
            constraint=models.UniqueConstraint(
                deferrable=django.db.models.constraints.Deferrable["DEFERRED"],
                fields=("platform_organization", "platform_user"),
                name="unique_overlap",
            ),
        ),
    ]



class DuplicationError(Exception):
    """
    Exception raised for issues that arise during duplication.

    Attributes:
        message(str): Explanation of the error.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



from contextlib import contextmanager
from unittest.mock import AsyncMock, patch

import async_property
import pytest
import rest_framework
from adrf import serializers as adrf_serializers
from django.db.models import Manager, Model, TextField


@pytest.fixture(scope="session")
def get_test_model():
    class SomeModel(Model):
        name = TextField()

        async def asave(*args, **kwargs):
            return True

    return SomeModel


@pytest.fixture(scope="session")
def get_test_audit_model():
    from smoothglue.core.models import AuditModel

    class SomeAuditModel(AuditModel):
        name = TextField()

        async def asave(*args, **kwargs):
            return True

    return SomeAuditModel


@pytest.fixture
def get_test_async_serializer(get_test_model):
    SomeModel = get_test_model

    class SomeAsyncSerializer(adrf_serializers.ModelSerializer):
        @async_property.async_property
        async def adata(self):
            return {}

        def create(self, *args, **kwargs):
            return SomeModel(id=1, name="some name")

        class Meta:
            model = SomeModel
            fields = ["name"]

    return SomeAsyncSerializer


@pytest.fixture
def get_test_sync_serializer(get_test_model):
    SomeModel = get_test_model

    class SomeSyncSerializer(rest_framework.serializers.ModelSerializer):
        def data(self):
            return {}

        def create(self, *args, **kwargs):
            return SomeModel(id=1, name="some name")

        class Meta:
            model = SomeModel
            fields = ["name"]

    return SomeSyncSerializer


@pytest.fixture
def patch_db_operations():
    @contextmanager
    def patch_db_operations_context(return_value: Model):
        with (
            patch.object(
                Manager, "create", autospec=True, return_value=return_value
            ) as mock_create,
            patch.object(
                Manager,
                "acreate",
                new_callable=AsyncMock,
                return_value=return_value,
            ) as mock_acreate,
            patch.object(
                Model, "save", autospec=True, return_value=return_value
            ) as mock_save,
        ):
            yield mock_create, mock_acreate, mock_save

    return patch_db_operations_context



from django.conf import settings
from django.db import models


class TimeAuditModel(models.Model):
    """Abstract class to give timestamp audit fields to other models"""

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class AuditModel(TimeAuditModel):
    """Abstract class to include user audits with timestamp audit fields"""

    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="%(class)ss_created",
        blank=True,
        null=True,
    )

    updated_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="%(class)ss_updated",
        blank=True,
        null=True,
    )

    class Meta:
        abstract = True


class Classification:
    """
    Classification data
    """

    def __init__(self, level, message):
        self.level = level
        self.message = message


class AppInfo:
    """
    Non DB object for serializing data
    """

    def __init__(self, classification) -> None:
        self.classification = classification



from rest_framework.renderers import BaseRenderer
import json
from django.core.serializers.json import DjangoJSONEncoder


class FixtureRenderer(BaseRenderer):
    media_type = "application/json"
    format = "fixture"
    charset = "utf-8"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if data is None:
            return ""

        if not isinstance(data, list):
            data = [data]

        # Get model from view
        model_name = "unknown"
        if renderer_context and "view" in renderer_context:
            view = renderer_context["view"]
            if hasattr(view, "queryset") and view.queryset is not None:
                model = view.queryset.model
                model_name = f"{model._meta.app_label}.{model._meta.model_name}"

        fixtures = []
        for item in data:
            fixture = {
                "model": model_name,
                "pk": item.get("id"),
                "fields": {k: v for k, v in item.items() if k != "id"},
            }
            fixtures.append(fixture)

        return json.dumps(fixtures, indent=2, cls=DjangoJSONEncoder)



# Generated by Django 4.2.20 on 2025-03-28 01:55

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0006_alter_organizationcategory_data_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='organizationcategory',
            options={'verbose_name_plural': 'Organization Categories'},
        ),
    ]



from unittest import mock

from django.shortcuts import reverse
from rest_framework.test import APITestCase

from smoothglue.authentication.models import PlatformOrganization, PlatformUser


class TestPlatformUser(APITestCase):
    ENDPOINT = reverse("users-list")

    def setUp(self):
        self.obj = PlatformUser.objects.create(
            username="Test",
            email="test@platform.test",
        )
        self.org = PlatformOrganization.objects.create(name="test")

    def test_get_list(self):
        resp = self.client.get(self.ENDPOINT, format="json")
        self.assertEqual(resp.status_code, 200)
        self.assertTrue(len(resp.data) > 0)

    def test_get(self):
        resp = self.client.get(f"{self.ENDPOINT}{self.obj.id}/", format="json")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data["username"], self.obj.username)

    @mock.patch(
        "smoothglue.authentication.permissions.UserModificationPermission.has_object_permission"
    )
    def test_patch(self, mockPermission):
        """
        PATCH should only update data that is passed in
        """
        mockPermission.return_value = True
        resp = self.client.patch(
            f"{self.ENDPOINT}{self.obj.id}/",
            data={
                "first_name": "Test",
            },
            format="json",
        )
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data["first_name"], "Test")

    @mock.patch(
        "smoothglue.authentication.permissions.UserModificationPermission.has_object_permission"
    )
    def test_put(self, mockPermission):
        """
        PUT update all data that is passed in
        """
        mockPermission.return_value = True
        resp = self.client.put(
            f"{self.ENDPOINT}{self.obj.id}/",
            data={
                "username": "TEST",
                "first_name": "John",
                "last_name": "Doe",
                "email": "test@test.com",
            },
            format="json",
        )
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data["username"], "TEST")
        self.assertEqual(resp.data["first_name"], "John")
        self.assertEqual(resp.data["last_name"], "Doe")

    @mock.patch(
        "smoothglue.authentication.permissions.UserModificationPermission.has_object_permission"
    )
    def test_post(self, mockPermission):
        mockPermission.return_value = True
        resp = self.client.post(
            self.ENDPOINT,
            data={
                "username": "TEST",
                "first_name": "John",
                "last_name": "Doe",
                "email": "test@test.com",
            },
            format="json",
        )
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp.data["username"], "TEST")
        self.assertEqual(resp.data["first_name"], "John")
        self.assertEqual(resp.data["last_name"], "Doe")

    @mock.patch(
        "smoothglue.authentication.permissions.UserModificationPermission.has_object_permission"
    )
    def test_delete(self, mockPermission):
        mockPermission.return_value = True
        resp = self.client.delete(f"{self.ENDPOINT}{self.obj.id}/", format="json")
        self.assertEqual(resp.status_code, 204)

    @mock.patch(
        "smoothglue.authentication.permissions.UserModificationPermission.has_object_permission"
    )
    def test_patch_nested_org(self, mockPermission):
        """
        PATCH should only update data that is passed in
        """
        mockPermission.return_value = True
        resp = self.client.patch(
            f"{self.ENDPOINT}{self.obj.id}/",
            data={"organizations": [str(self.org.id)]},
            format="json",
        )
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data["organizations"], [self.org.id])



import uuid

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser, Group
from django.db import models
from django.db.models.constraints import Deferrable, UniqueConstraint
from django.db.models.signals import m2m_changed
from django.dispatch.dispatcher import receiver

from smoothglue.core.models import AuditModel


class PlatformUser(AbstractUser):
    """
    Extended Django User Model that has additional
    properties that can be stored against the user
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    data = models.JSONField(default=dict, blank=True)
    email = models.EmailField(unique=True)
    organizations = models.ManyToManyField(
        to="PlatformOrganization",
        default=None,
        through="OrganizationMember",
        related_name="members",
        through_fields=["platform_user", "platform_organization"],
    )
    user_image = models.ImageField(upload_to=None, null=True, blank=True)

    def __str__(self):
        full_name = f"{self.first_name} {self.last_name}".strip()
        return full_name if full_name else str(self.username)


@receiver(models.signals.post_save, sender=get_user_model())
def post_save_user_signal_handler(instance, created, **kwargs):
    """
    Allows newly created users to be added to the global user group if the
    GLOBAL_USER_GROUP_NAME configuration is a set and the group exist in the
    database.
    """
    if created:
        global_user_group = "Participant"
        if hasattr(settings, "GLOBAL_USER_GROUP_NAME"):
            global_user_group = settings.GLOBAL_USER_GROUP_NAME
        group = Group.objects.filter(name=global_user_group).first()
        if group:
            instance.groups.add(group)
            instance.save()


class OrganizationMember(AuditModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    platform_organization = models.ForeignKey(
        "PlatformOrganization", on_delete=models.CASCADE
    )
    platform_user = models.ForeignKey(PlatformUser, on_delete=models.CASCADE)

    class Meta:
        constraints = [
            UniqueConstraint(
                fields=["platform_organization", "platform_user"],
                name="unique_overlap",
                deferrable=Deferrable.DEFERRED,
            ),
        ]


class OrganizationCategory(AuditModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255, null=False, blank=False, unique=True)
    description = models.TextField(null=True, blank=True)
    data = models.JSONField(default=dict, blank=True)

    class Meta:
        verbose_name_plural = "Organization Categories"

    def __str__(self):
        return self.name


class PlatformOrganization(AuditModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255, null=False, blank=False, unique=True)
    group = models.ForeignKey(Group, on_delete=models.CASCADE, null=True, blank=True)
    parent_organization = models.ForeignKey(
        "PlatformOrganization",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        related_name="child_organizations",
        help_text="Parent Organization",
    )
    platform_users = models.ManyToManyField(
        to=PlatformUser,
        default=None,
        through=OrganizationMember,
        through_fields=["platform_organization", "platform_user"],
    )
    abbreviation = models.CharField(max_length=10, null=True, blank=True)
    org_category = models.ForeignKey(
        OrganizationCategory, on_delete=models.CASCADE, null=True, blank=True
    )
    data = models.JSONField(default=dict, blank=True)

    def __str__(self) -> str:
        return self.name


def is_group_security_enabled():
    enable_group_security = False
    if hasattr(settings, "ENABLE_GROUP_SECURITY"):
        enable_group_security = settings.ENABLE_GROUP_SECURITY
    return enable_group_security


@receiver(models.signals.post_save, sender=PlatformOrganization)
def post_save_org_signal_handler(instance, created, **kwargs):
    """
    when ENABLE_GROUP_SECURITY is configured, a Django group object will be created
    anytime a new PlatformOrganization is created.
    """
    if created and is_group_security_enabled():
        group = Group.objects.create(name=instance.name)
        instance.group = group
        instance.save()


@receiver(models.signals.post_save, sender=OrganizationMember)
def post_save_org_member_signal_handler(instance, created, **kwargs):
    """
    when ENABLE_GROUP_SECURITY is configured, a Django group object will be created
    anytime a new PlatformOrganization is created.
    """
    if created and is_group_security_enabled() and instance.platform_organization.group:
        instance.platform_user.groups.add(instance.platform_organization.group)


@receiver(models.signals.post_delete, sender=OrganizationMember)
def post_delete_org_member_signal_handler(instance, **kwargs):
    """
    when ENABLE_GROUP_SECURITY is configured, delete user group membership when the user
    is taken out from the organization.
    """
    if is_group_security_enabled() and instance.platform_organization.group:
        instance.platform_user.groups.remove(instance.platform_organization.group)


@receiver(m2m_changed, sender=PlatformOrganization.platform_users.through)
def org_users_changed(sender, instance, action, pk_set=None, model=None, **kwargs):
    if isinstance(instance, PlatformOrganization):
        new_users = model.objects.filter(id__in=pk_set)
        for user in new_users:
            if action == "post_add":
                user.groups.add(instance.group)
            elif action == "post_remove":
                user.groups.remove(instance.group)
    elif isinstance(instance, PlatformUser):
        user = instance
        teams = model.objects.filter(id__in=pk_set)
        for team in teams:
            if action == "post_add":
                user.groups.add(team.group)
            elif action == "post_remove":
                user.groups.remove(team.group)



"""Tests for admin.py to improve coverage."""

from unittest.mock import MagicMock, patch

import pytest
from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.db import models

from smoothglue.core.admin import BaseAuditModelAdmin

User = get_user_model()


# Test model for admin testing
class AuditTestModel(models.Model):
    """Test model with audit fields."""

    name = models.CharField(max_length=100)
    created_by = models.ForeignKey(
        User, on_delete=models.CASCADE, related_name="+", null=True
    )
    updated_by = models.ForeignKey(
        User, on_delete=models.CASCADE, related_name="+", null=True
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        app_label = "core"


class AuditTestModelAdmin(BaseAuditModelAdmin):
    """Test admin class for AuditTestModel."""

    pass


@pytest.mark.django_db
class TestBaseAuditModelAdmin:
    """Test BaseAuditModelAdmin save methods."""

    def setup_method(self):
        """Set up test fixtures."""
        self.site = AdminSite()
        self.admin = AuditTestModelAdmin(AuditTestModel, self.site)
        self.user = User.objects.create_user(
            username="testuser", email="test@example.com"
        )

    def test_save_model_create(self):
        """Test save_model method when creating new object."""
        request = MagicMock()
        request.user = self.user

        obj = AuditTestModel(name="Test Object")
        form = MagicMock()
        change = False  # Creating new object

        # Mock the parent save_model method
        with patch.object(BaseAuditModelAdmin.__bases__[0], "save_model") as mock_super:
            self.admin.save_model(request, obj, form, change)

            # Check that audit fields are set correctly
            assert obj.created_by == self.user
            assert obj.updated_by == self.user
            mock_super.assert_called_once_with(request, obj, form, change)

    def test_save_model_update(self):
        """Test save_model method when updating existing object."""
        request = MagicMock()
        request.user = self.user

        # Create another user for original created_by
        original_user = User.objects.create_user(
            username="original", email="original@example.com"
        )

        obj = AuditTestModel(name="Test Object", created_by=original_user)
        form = MagicMock()
        change = True  # Updating existing object

        # Mock the parent save_model method
        with patch.object(BaseAuditModelAdmin.__bases__[0], "save_model") as mock_super:
            self.admin.save_model(request, obj, form, change)

            # Check that only updated_by is changed, not created_by
            assert obj.created_by == original_user  # Should remain unchanged
            assert obj.updated_by == self.user
            mock_super.assert_called_once_with(request, obj, form, change)

    def test_save_formset(self):
        """Test save_formset method."""
        request = MagicMock()
        request.user = self.user
        form = MagicMock()

        # Create mock formset
        formset = MagicMock()

        # Create test instances - one without created_by, one with created_by
        new_instance = MagicMock()
        new_instance.created_by = None  # Simulate new instance
        new_instance.save = MagicMock()

        existing_instance = MagicMock()
        existing_instance.created_by = self.user  # Simulate existing instance
        existing_instance.save = MagicMock()

        # Mock hasattr to return False for new instance, True for existing
        def mock_hasattr(obj, attr):
            if attr == "created_by":
                return obj.created_by is not None
            return True

        formset.save.return_value = [new_instance, existing_instance]

        change = True

        with patch("builtins.hasattr", side_effect=mock_hasattr):
            self.admin.save_formset(request, form, formset, change)

        # Check that new instance gets created_by set
        assert new_instance.created_by == self.user
        assert new_instance.updated_by == self.user

        # Check that existing instance only gets updated_by set
        assert existing_instance.updated_by == self.user

        # Verify save was called on instances
        new_instance.save.assert_called_once()
        existing_instance.save.assert_called_once()

        # Verify formset methods were called
        formset.save.assert_called_once_with(commit=False)



from django.urls import include, path
from rest_framework import routers
from rest_framework.schemas import get_schema_view

from smoothglue.tracker.views import APIChangeLogViewSet, AppErrorLogViewSet

ROUTER = routers.DefaultRouter()

ROUTER.register(r"apilog", APIChangeLogViewSet, basename="apichangelog")
ROUTER.register(r"errorlogger", AppErrorLogViewSet, basename="apperrorlogger")

schema_view = get_schema_view(
    title="SmoothGlue Tracker",
    description="SmoothGlue Tracker API",
    version="1.0.0",
    patterns=[
        path("", include("smoothglue.tracker.urls")),
    ],
)

urlpatterns = [
    path("", include(ROUTER.urls)),
]



import json
import logging

from django.db import connection


class ChangeLogHandler(logging.Handler):
    """
    Handles any changes to a model by the API.
    """

    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record):
        msg = record.msg
        request = msg["request"]
        # Only record changes
        if request["method"] in ["POST", "PUT", "PATCH", "DELETE"]:
            username = "Unknown"
            if msg.get("user"):
                username = msg["user"].get("username", "Unknown")

            data = request["data"]
            if data:
                # Attempt to remove middlewaretoken
                parsed_data = json.loads(request["data"])
                parsed_data.pop("csrfmiddlewaretoken", None)
                data = json.dumps(parsed_data)
            with connection.cursor() as cursor:
                cursor.execute(
                    """
                    INSERT INTO smoothglue_tracker_apichangelog (username, full_path,
                                                    method, timestamp,
                                                    data, params)
                    VALUES (%s, %s, %s, %s, %s, %s)
                    """,
                    [
                        str(username)[0:200],
                        str(request["full_path"])[0:200],
                        str(request["method"])[0:30],
                        msg["timestamp"],
                        str(data),
                        str(request["query_params"]),
                    ],
                )



# Generated by Django 4.2.15 on 2024-09-06 15:53

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("authentication", "0002_organizationmember_platformorganization_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="platformuser",
            name="organizations",
            field=models.ManyToManyField(
                default=None,
                related_name="members",
                through="authentication.OrganizationMember",
                to="authentication.platformorganization",
            ),
        ),
        migrations.AlterField(
            model_name="organizationmember",
            name="platform_organization",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="authentication.platformorganization",
            ),
        ),
        migrations.AlterField(
            model_name="organizationmember",
            name="platform_user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AlterField(
            model_name="platformuser",
            name="email",
            field=models.EmailField(max_length=254, unique=True),
        ),
    ]



import json
import logging
import unittest
from unittest.mock import patch

from smoothglue.tracker.logging import ChangeLogHandler


class TestChangeLogHandler(unittest.TestCase):
    def setUp(self):
        self.handler = ChangeLogHandler()
        self.mock_record = logging.LogRecord(
            name="test-logger",
            level=logging.INFO,
            pathname="test_pathname",
            lineno=1,
            msg={},
            args=(),
            exc_info=None,
            func="test_func",
        )

    def _prepare_record_msg(
        self, method, user, data, full_path, query_params, timestamp
    ):
        """Helper to construct the msg dictionary for the log record."""
        self.mock_record.msg = {
            "request": {
                "method": method,
                "data": data,
                "full_path": full_path,
                "query_params": query_params,
            },
            "user": user,
            "timestamp": timestamp,
        }

    @patch("smoothglue.tracker.logging.connection")
    def test_emit_non_change_method_does_not_log(self, mock_connection):
        """Test that methods like GET don't result in a database write."""
        mock_cursor = mock_connection.cursor.return_value.__enter__.return_value

        self._prepare_record_msg(
            method="GET",
            user={"username": "testuser"},
            data=None,
            full_path="/api/items",
            query_params="{}",
            timestamp="2023-10-27T10:00:00Z",
        )
        self.handler.emit(self.mock_record)
        mock_cursor.execute.assert_not_called()

    @patch("smoothglue.tracker.logging.connection")
    def test_emit_post_request_with_user_and_data(self, mock_connection):
        """Test a POST request with user and data, including CSRF token removal."""
        mock_cursor = mock_connection.cursor.return_value.__enter__.return_value
        timestamp = "2023-10-27T10:00:00Z"
        original_data_dict = {"key": "value", "csrfmiddlewaretoken": "testtoken"}
        original_data_json = json.dumps(original_data_dict)
        expected_data_json = json.dumps({"key": "value"})
        username = "testuser"
        full_path = "/api/resource"
        query_params_str = "{'param': 'value'}"

        self._prepare_record_msg(
            method="POST",
            user={"username": username},
            data=original_data_json,
            full_path=full_path,
            query_params=query_params_str,
            timestamp=timestamp,
        )
        self.handler.emit(self.mock_record)

        expected_sql = """
                    INSERT INTO smoothglue_tracker_apichangelog (username, full_path,
                                                    method, timestamp,
                                                    data, params)
                    VALUES (%s, %s, %s, %s, %s, %s)
                    """
        expected_params = [
            username,
            full_path,
            "POST",
            timestamp,
            expected_data_json,
            query_params_str,
        ]
        mock_cursor.execute.assert_called_once()
        call_args = mock_cursor.execute.call_args[0]
        self.assertEqual(" ".join(expected_sql.split()), " ".join(call_args[0].split()))
        self.assertEqual(expected_params, call_args[1])

    @patch("smoothglue.tracker.logging.connection")
    def test_emit_put_request_anonymous_user_no_data(self, mock_connection):
        """Test a PUT request with an anonymous user and no data."""
        mock_cursor = mock_connection.cursor.return_value.__enter__.return_value
        timestamp = "2023-10-27T10:05:00Z"
        full_path = "/api/resource/1"
        query_params_str = "{}"

        self._prepare_record_msg(
            method="PUT",
            user=None,  # Anonymous user
            data=None,  # No data
            full_path=full_path,
            query_params=query_params_str,
            timestamp=timestamp,
        )
        self.handler.emit(self.mock_record)

        expected_params = [
            "Unknown",
            full_path,
            "PUT",
            timestamp,
            "None",
            query_params_str,
        ]
        mock_cursor.execute.assert_called_once()
        self.assertEqual(expected_params, mock_cursor.execute.call_args[0][1])

    @patch("smoothglue.tracker.logging.connection")
    def test_emit_delete_request_data_without_csrf_token(self, mock_connection):
        """Test DELETE with data that doesn't contain a CSRF token."""
        mock_cursor = mock_connection.cursor.return_value.__enter__.return_value
        timestamp = "2023-10-27T10:10:00Z"
        data_dict = {"id": 123, "force": True}
        data_json = json.dumps(data_dict)
        username = "deleter_user"
        full_path = "/api/items/123"
        query_params_str = "{'hard_delete': 'true'}"

        self._prepare_record_msg(
            method="DELETE",
            user={"username": username},
            data=data_json,
            full_path=full_path,
            query_params=query_params_str,
            timestamp=timestamp,
        )
        self.handler.emit(self.mock_record)

        expected_params = [
            username,
            full_path,
            "DELETE",
            timestamp,
            data_json,
            query_params_str,
        ]
        mock_cursor.execute.assert_called_once()
        self.assertEqual(expected_params, mock_cursor.execute.call_args[0][1])

    @patch("smoothglue.tracker.logging.connection")
    def test_emit_patch_request_empty_json_object_data(self, mock_connection):
        """Test PATCH with data being an empty JSON object string '{}'."""
        mock_cursor = mock_connection.cursor.return_value.__enter__.return_value
        timestamp = "2023-10-27T10:15:00Z"
        empty_json_data_str = "{}"
        username = "patch_user_empty_json"
        full_path = "/api/config/def"
        query_params_str = "{}"

        self._prepare_record_msg(
            method="PATCH",
            user={"username": username},
            data=empty_json_data_str,
            full_path=full_path,
            query_params=query_params_str,
            timestamp=timestamp,
        )
        self.handler.emit(self.mock_record)

        expected_params = [
            username,
            full_path,
            "PATCH",
            timestamp,
            empty_json_data_str,
            query_params_str,
        ]
        mock_cursor.execute.assert_called_once()
        self.assertEqual(expected_params, mock_cursor.execute.call_args[0][1])

    @patch("smoothglue.tracker.logging.connection")
    def test_emit_data_truncation(self, mock_connection):
        """Test that username and full_path are truncated if too long."""
        mock_cursor = mock_connection.cursor.return_value.__enter__.return_value
        timestamp = "2023-10-27T10:20:00Z"
        long_username = "u" * 250
        long_full_path = "/a/b/c/" * 100
        method = "POST"

        self._prepare_record_msg(
            method=method,
            user={"username": long_username},
            data=None,
            full_path=long_full_path,
            query_params="{}",
            timestamp=timestamp,
        )
        self.handler.emit(self.mock_record)

        expected_params = [
            long_username[:200],
            long_full_path[:200],
            method,
            timestamp,
            "None",
            "{}",
        ]
        mock_cursor.execute.assert_called_once()
        actual_params = mock_cursor.execute.call_args[0][1]

        self.assertEqual(expected_params[0], actual_params[0])
        self.assertEqual(len(actual_params[0]), 200)
        self.assertEqual(expected_params[1], actual_params[1])
        self.assertEqual(len(actual_params[1]), 200)
        self.assertEqual(expected_params[2:], actual_params[2:])

    @patch("smoothglue.tracker.logging.connection")
    def test_emit_user_is_none(self, mock_connection):
        """Test when msg['user'] itself is None, not just missing 'username'."""
        mock_cursor = mock_connection.cursor.return_value.__enter__.return_value
        timestamp = "2023-10-27T10:25:00Z"

        self._prepare_record_msg(
            method="POST",
            user=None,
            data='{"key": "value"}',
            full_path="/api/somepath",
            query_params="{}",
            timestamp=timestamp,
        )
        self.handler.emit(self.mock_record)

        expected_params_subset = [
            "Unknown",
            "/api/somepath",
            "POST",
            timestamp,
            '{"key": "value"}',
            "{}",
        ]
        mock_cursor.execute.assert_called_once()
        self.assertEqual(expected_params_subset, mock_cursor.execute.call_args[0][1])

    @patch("smoothglue.tracker.logging.connection")
    def test_emit_user_dict_empty_or_no_username(self, mock_connection):
        """Test when msg['user'] is a dict but empty or lacks 'username' key."""
        mock_cursor = mock_connection.cursor.return_value.__enter__.return_value
        timestamp = "2023-10-27T10:30:00Z"

        self._prepare_record_msg(
            method="PUT",
            user={},
            data=None,
            full_path="/api/path1",
            query_params="{}",
            timestamp=timestamp,
        )
        self.handler.emit(self.mock_record)
        mock_cursor.execute.assert_called_once()
        self.assertEqual("Unknown", mock_cursor.execute.call_args[0][1][0])
        mock_cursor.reset_mock()

        self._prepare_record_msg(
            method="PATCH",
            user={"id": 1, "email": "user@example.com"},
            data=None,
            full_path="/api/path2",
            query_params="{}",
            timestamp=timestamp,
        )
        self.handler.emit(self.mock_record)
        mock_cursor.execute.assert_called_once()
        self.assertEqual("Unknown", mock_cursor.execute.call_args[0][1][0])



from django.conf import settings
from django.test import TestCase

from smoothglue.authentication.models import (
    OrganizationMember,
    PlatformOrganization,
    PlatformUser,
)


class TestPlatformOrganization(TestCase):
    def test_create_org_should_create_group(self):
        settings.ENABLE_GROUP_SECURITY = True
        name = "Org1"
        PlatformOrganization.objects.create(
            name=name,
        )
        obj = PlatformOrganization.objects.first()
        self.assertTrue(obj.pk)
        self.assertEqual(name, obj.name)
        self.assertIsNotNone(obj.group)

    def test_create_org_should_not_create_group_when_enable_group_security_is_false(
        self,
    ):
        settings.ENABLE_GROUP_SECURITY = False
        name = "Org1"
        PlatformOrganization.objects.create(
            name=name,
        )
        obj = PlatformOrganization.objects.first()
        self.assertTrue(obj.pk)
        self.assertEqual(name, obj.name)
        self.assertIsNone(obj.group)

    def test_user_org_member_should_have_group_membership(self):
        settings.ENABLE_GROUP_SECURITY = True
        name = "Org1"
        PlatformOrganization.objects.create(
            name=name,
        )
        obj = PlatformOrganization.objects.first()
        self.assertTrue(obj.pk)
        self.assertEqual(name, obj.name)
        self.assertIsNotNone(obj.group)

        user = PlatformUser.objects.create(username="userA")
        obj.platform_users.add(user)

        self.assertTrue(user.groups.contains(obj.group))

        obj.platform_users.remove(user)

        self.assertFalse(user.groups.contains(obj.group))

    def test_user_org_member_create_should_add_group_membership(self):
        settings.ENABLE_GROUP_SECURITY = True
        name = "Org1"
        org = PlatformOrganization.objects.create(
            name=name,
        )
        user = PlatformUser.objects.create(username="userA")

        OrganizationMember.objects.create(platform_user=user, platform_organization=org)

        user = PlatformUser.objects.filter(username="userA").first()

        self.assertTrue(len(user.groups.all()), 1)



import logging

from django.conf import settings
from django.db import transaction
from drf_spectacular.utils import OpenApiParameter, extend_schema
from rest_framework import status, viewsets
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.schemas.openapi import AutoSchema
from rest_framework.views import exception_handler as default_exception_handler

from smoothglue.core.models import AppInfo, Classification
from smoothglue.core.serializers.app_info import AppInfoSerializer
from smoothglue.tracker.utils import create_change_log

logger = logging.getLogger(__name__)


class ListModelViewSet(viewsets.ModelViewSet):
    """
    A viewset that extends ModelViewSet to handle the creation of lists of objects.
    It includes enhanced logging for change-making operations and uses atomic
    transactions where necessary to ensure data integrity.
    """

    @extend_schema(
        methods=["get"],
        parameters=[
            OpenApiParameter(
                name="fields",
                type={"type": "array", "items": {"type": "string"}},
                location=OpenApiParameter.QUERY,
                description="Fields to include in the response",
            ),
        ],
    )
    def list(self, request, *args, **kwargs):
        return super().list(request, *args, **kwargs)

    def get_exception_handler(self):
        return self.exception_handler

    @staticmethod
    def exception_handler(exc, context):
        response = default_exception_handler(exc, context)

        if response is not None:
            return response

        error_verbs = {
            "DELETE": "deleting",
            "PUT": "modifying",
            "PATCH": "modifying",
            "POST": "creating",
        }
        logger.error(
            "Error %s object - %s: %s",
            error_verbs.get(context["request"].method, ""),
            type(exc).__name__,
            str(exc),
            exc_info=True,
        )
        return Response({"error": str(exc)}, status=status.HTTP_400_BAD_REQUEST)

    @transaction.atomic
    def create(self, request, *args, **kwargs):
        """
        Creates a single object or a list of objects. All operations are wrapped in an atomic
        transaction to ensure data integrity. A change log is created after the operation.
        """
        is_many = isinstance(request.data, list)
        serializer = self.get_serializer(data=request.data, many=is_many)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        create_change_log(request)
        headers = self.get_success_headers(serializer.data)
        return Response(
            serializer.data, status=status.HTTP_201_CREATED, headers=headers
        )

    def update(self, request, *args, **kwargs):
        """
        Updates an existing object. A change log is created after the operation.
        """
        response = super().update(request, *args, **kwargs)
        create_change_log(request)
        return response

    def partial_update(self, request, *args, **kwargs):
        """
        Partially updates an existing object. A change log is created after the operation.
        """
        response = super().partial_update(request, *args, **kwargs)
        create_change_log(request)
        return response

    @transaction.atomic
    def destroy(self, request, *args, **kwargs):
        """
        Deletes an object. The operation is wrapped in an atomic transaction to ensure
        data integrity. A change log is created after the operation.
        """
        response = super().destroy(request, *args, **kwargs)
        create_change_log(request)
        return response


class CustomDataSchema(AutoSchema):
    """
    Overrides the JSONField data field with explicit data schema
    """

    def __init__(self, **kwargs):
        self.data_schema = kwargs.pop("data_schema")
        super().__init__(**kwargs)

    def map_field(self, field):
        if field.field_name == "data":
            return self.data_schema
        return super().map_field(field)


@extend_schema(responses=AppInfoSerializer)
@api_view(http_method_names=["GET"])
def get_app_info(request):
    """
    Retrieve App Info with classification configuration for the application
    """
    classification_level = settings.CLASSIFICATION_LEVEL.lower().strip()
    info = AppInfo(
        Classification(
            level=classification_level, message=settings.CLASSIFICATION_MESSAGE
        )
    )
    return Response(AppInfoSerializer(instance=info).data)



from django.shortcuts import get_object_or_404
from django_filters.rest_framework import DjangoFilterBackend
from drf_spectacular.utils import extend_schema
from rest_framework import status
from rest_framework.decorators import api_view
from rest_framework.response import Response

from smoothglue.authentication.permissions import UserModificationPermission
from smoothglue.core.views import ListModelViewSet

from .models import (
    OrganizationCategory,
    OrganizationMember,
    PlatformOrganization,
    PlatformUser,
)
from .serializers import (
    OrganizationCategorySerializer,
    OrganizationMemberSerializer,
    PlatformOrganizationSerializer,
    PlatformUserSerializer,
)


@extend_schema(
    operation_id="authentication_active_user_retrieve",
    responses={200: PlatformUserSerializer},
    tags=["authentication"],
    description="Retrieve the active user's information.",
)
@api_view(
    ["GET"]
)  # Required for DRF Spectacular to detect function-based views in OpenAPI schema
def get_active_user(request):
    """
    Retrieve the active user's information.

    Responds with the serialized details of the authenticated user,
    or a 404 response if no matching user is found.
    """

    # Retrieve the user and serialize their data
    platform_user = get_object_or_404(PlatformUser, id=request.user.id)
    serialized_user = PlatformUserSerializer(platform_user)

    return Response(serialized_user.data, status=status.HTTP_200_OK)


class PlatformUserViewSet(ListModelViewSet):
    queryset = PlatformUser.objects.all()
    serializer_class = PlatformUserSerializer
    permission_classes = [UserModificationPermission]


@extend_schema(
    operation_id="authentication_headers_retrieve",
    responses={200: dict},
    tags=["authentication"],
    description="Retrieve request headers for debugging.",
)
@api_view(
    ["GET"]
)  # Required for DRF Spectacular to detect function-based views in OpenAPI schema
def what_are_my_headers(request):
    return Response(dict(request.headers), status=200)


class PlatformOrganizationViewSet(ListModelViewSet):
    filter_backends = [DjangoFilterBackend]
    filterset_fields = {"parent_organization": ["exact", "isnull"]}
    queryset = PlatformOrganization.objects.all()
    serializer_class = PlatformOrganizationSerializer


class OrganizationMemberViewSet(ListModelViewSet):
    filter_backends = [DjangoFilterBackend]
    queryset = OrganizationMember.objects.all()
    filterset_fields = {
        "platform_organization": ["exact", "in"],
        "platform_user": ["exact", "in"],
    }
    serializer_class = OrganizationMemberSerializer
    http_method_names = ["get", "head"]


class OrganizationCategoryViewSet(ListModelViewSet):
    filter_backends = [DjangoFilterBackend]
    queryset = OrganizationCategory.objects.all()
    serializer_class = OrganizationCategorySerializer



# Generated by Django 4.2.7 on 2023-11-16 14:11

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="APIChangeLog",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("username", models.CharField(blank=True, max_length=200)),
                ("full_path", models.CharField(blank=True, max_length=200)),
                ("method", models.CharField(blank=True, max_length=30)),
                ("timestamp", models.DateTimeField(blank=True)),
                ("data", models.JSONField(default=dict)),
                ("params", models.JSONField(default=dict)),
            ],
            options={
                "verbose_name": "API Change Log",
            },
        ),
        migrations.CreateModel(
            name="AppErrorLog",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "level",
                    models.CharField(
                        choices=[
                            ("WARNING", "Warning"),
                            ("ERROR", "Error"),
                            ("FATAL", "Fatal"),
                        ],
                        max_length=10,
                    ),
                ),
                ("message", models.CharField(max_length=200)),
                (
                    "stack_trace",
                    models.TextField(
                        blank=True,
                        help_text="Stack trace error logging if it's available",
                        null=True,
                    ),
                ),
                (
                    "client_data",
                    models.TextField(
                        blank=True,
                        help_text="Any other client data, user agent, ip address etc",
                        null=True,
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        help_text="User data if available",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "App Error Log",
            },
        ),
    ]



from django.urls import include, path
from rest_framework import routers

import smoothglue.authentication.urls
import smoothglue.tracker.urls

router = routers.DefaultRouter()

router_registries = [
    smoothglue.tracker.urls.ROUTER.registry,
    smoothglue.authentication.urls.ROUTER.registry,
]

for router_registry in router_registries:
    router.registry.extend(router_registry)

urlpatterns = [
    path("", include(router.urls)),
]



from rest_framework import permissions


class UserModificationPermission(permissions.BasePermission):
    """
    Custom permission to only allow admins or the user themselves to edit their information.
    """

    def has_object_permission(self, request, view, obj):
        # Read permissions are allowed to any request,
        # so we'll always allow GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            return True

        # Write permissions are only allowed to the user of the profile or admin
        return obj == request.user or request.user.is_staff
